
#ifndef __B_CONNECTED_COMPONENTS_H
#define __B_CONNECTED_COMPONENTS_H

/*
CLASS
  BConnectedComponents

  Template defines a basical interface for the UnionList class to be used to
  find the Bi-ConnectedComponents of a set of elementT type elements.
  A Bi Connected Component is a component of a graph in which every two 
  vertices can be connected using two vertices-disjoint routes.

  To achieve that the class use the BComponent class with the GraphNode
  wrapper around the elements, and a DFS search on the graph. The components
  are created using the low function computed on-the-fly while the DFS walk on 
  the graph.

KEYWORDS
  ConnectedComponents BConnectedComponents, set, Component, ComponentNode, 
  BComponent, GraphNode

AUTHORS
  Ram Nathaniel <ramn@math.tau.ac.il>

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1998.

GOALS
  Create an easy interface to the UnionList engine for the purpose of finding
  the BiConnectedComponents of a given set. 

USAGE
  The user should inherit from this class redifinning the distance function
  Dist. After that the use of the FindBCC method will return a UnionList of 
  BComponent objects (castable to a vector).

  EXAMPLE
  class MyBCC:public BConnectedComponents<elements>{
  public:
     float Dist(element& i, element& j, 
     MoleculeBase& model, MoleculeBase& scene){
        calculate distance here
        }
     };

  void main(){
     .
     .
     .
     UnionList<int, BComponent <element> > *component=
       MyBCC.FindBCC(model, scene, vector<element>, MaxDist);

     UnionList<int, BComponent <element> >::iterator it;
     for (it = component->begin(); it != component->end(); ++it)
       (*it).Print();
     }
  END
  
  Note that the distance function is not predeclared.
  
  */


#include <stl.h>
#include "UnionList.h"
#include "../gamb++/MoleculeBase.h"

#include "BComponent.h"
#include "Connected.h"

#define ResUnion  UnionList<int, BComponent<class elementT> >

template <class elementT> class BConnectedComponents:
  public ConnectedComponents<class elementT>{
public:
    //main function - see example for typical usage form.
    ResUnion *FindBCC(MoleculeBase& model,
                      MoleculeBase& scene,
                      vector<class elementT>& elements,
                      float MaxDist);
    
    //standard DFS walk on the graph, calculating the low function on 
    //the vertices on-the-fly.
    void DFS(int *time, DFSGraphNode<class elementT> *n, 
             DFSGraphNode<class elementT> *parent);
    
    //creates the BConnectedComponents vector after the DFS runs on the graph.
    void MakeBCC(DFSGraphNode<elementT> *n, 
                 DFSGraphNode<elementT> *parent, 
                 ResUnion *Componenents);
};

template <class elementT>
void BConnectedComponents<class elementT>::
DFS(int *time, DFSGraphNode<class elementT> *n, 
    DFSGraphNode<class elementT> *parent){

  int min_low=*time;
  
  n->dfs_time=*time;
  n->dfs_parent=parent;
  
  //go through all neighbours except the parent.
  for (unsigned int i=0; i < n->Neighbours.size(); ++i){
    DFSGraphNode<elementT> *ni=(DFSGraphNode<elementT>*)n->Neighbours[i];
    if (ni != parent){
      if (ni->dfs_time != -1){//if not marked
        (*time)++;
        DFS(time, ni, n);
        if (ni->low < min_low) min_low=ni->low;
      }
      else{ //else just update low if needed.
        if (ni->dfs_time < min_low) 
          min_low=ni->dfs_time;
      }
    }
  }
  n->low=min_low;
}

template <class elementT>
void BConnectedComponents <class elementT>::
MakeBCC(DFSGraphNode<elementT> *n, DFSGraphNode<elementT> *parent, 
        ResUnion *Components){

  UnionList<int, BComponent<elementT> >::iterator i_it, j_it;

  //go through all neighbours except the parent.
  for (unsigned int i=0; i < (n->Neighbours).size(); ++i){
    DFSGraphNode<elementT>* other=(DFSGraphNode<elementT>*)n->Neighbours[i];
    if (other != parent){
      if (n->dfs_time >= other->low){
        cout << "joining " << n->element_index 
             << " and " << other->element_index << endl;
        i_it=Components->getUnion(n->element_index);
        j_it=Components->getUnion(other->element_index);
        if (i_it != j_it){
          (*i_it).join(*j_it);
          Components->unite(n->element_index, other->element_index);
        }
      }
    }
  }
}

template <class elementT>
ResUnion * BConnectedComponents<class elementT>::
FindBCC(MoleculeBase& model, MoleculeBase& scene,
        vector<class elementT>& elements, float MaxDist){

  cout << "inside FindBCC" << endl;

  ResUnion *Components;
  Components=new UnionList<int, BComponent<elementT> >;
  vector<DFSGraphNode<elementT> * > Nodes;
  
  //create unions - this phase is nessecary because I use Component::join
  //before I use the unite method so the unions are assumed to exist at 
  //that stage.
  for (int i=0; i < (int)elements.size(); ++i){
    elementT& e=elements[i];
    //create node for the element.
    DFSGraphNode<elementT> *n = new DFSGraphNode<elementT>(&e, i);
    Nodes.push_back(n);
    
    //create component for the node.
    BComponent<elementT> *c = 
      new BComponent<elementT>((DFSGraphNode<elementT>*)n);
    Components->addUnion(i, *c);
  }

  cout << "Unions created." << endl;
  
  //create the edges of the graph.
  for (int i=0; i < (int)elements.size(); ++i){
    for (int j=i+1; j < (int)elements.size(); ++j){
      if (Dist(elements[i], elements[j], model, scene) < MaxDist){
        //the two Nodes should be connected.
        Nodes[i]->ConnectTo(Nodes[j]);
        Nodes[j]->ConnectTo(Nodes[i]);
        cout << "Edge created : " << i << " - " << j << endl;
      }
    }
  }
  
  cout << "Edges Created." << endl;
  
  for (int i=0; i < (int)elements.size(); ++i){
    cout << "Node : " << i << " Neighbours : " 
         << Nodes[i]->Neighbours.size() << endl;
  }

  vector<DFSGraphNode<elementT> *> dfs_roots;
      //start the DFS
  int counter=0;
  while (counter < (int)Nodes.size()){
    unsigned int index=0;
    //find the smallest indexed Node which wasn't touched yet.
    //this is very slow (can be O(n^2)) but the best I found.
    for (index=0; 
         (index < Nodes.size()) && (Nodes[index]->dfs_time != -1) ;
         ++index);
    //if any nodes left process them.
    dfs_roots.push_back(Nodes[index]);
    if (index < Nodes.size()){
      cout << "Entering DFS" << endl;
      cout << "counter is " << counter 
           << " index is " << index << endl;
      DFS(&counter, Nodes[index], NULL);
    }
    else break;
  }
  
  cout << "DFS done. counter = " 
       << counter << " and Nodes.size() = "
       << Nodes.size() << endl;

  cout << "Number of dfs roots : " << dfs_roots.size() << endl;

  //unite to BCCs on each subtree of the dfs result.
  for (unsigned int i=0; i < dfs_roots.size(); ++i){
    MakeBCC(dfs_roots[i], NULL, Components);
  }
  
  cout << "Found BCCs." << endl;

  return Components;
}

#endif








