#ifndef __CONNECTED_COMPONENTS_H
#define __CONNECTED_COMPONENTS_H
/*
CLASS
  ConnectedComponents

  Template defines a basical interface for the UnionList class to be used to
  find the ConnectedComponents of a set of elementT type elements.

KEYWORDS
  Connected Components, set, Component, Node

AUTHORS
  Ram Nathaniel <ramn@math.tau.ac.il>

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1998.

GOALS
  Create an easy interface to the UnionList engine for the purpose of finding
  the ConnectedComponents of a given set. 

USAGE
  The user should inherit from this class redifinning the distance function
  Dist. After that the use of the FindCC method will return a UnionList of 
  Component objects (castable to a vector).

  EXAMPLE
  class MyCC:public ConnectedComponents<elements>{
  public:
     float Dist(element& i, element& j, 
     MoleculeBase& model, MoleculeBase& scene){
        calculate distance here
        }
     };

  void main(){
     .
     .
     .
     UnionList<int, Component <element> > *component=
       MyCC.FindCC(model, scene, vector<element>, MaxDist);

     UnionList<int, Component <element> >::iterator it;
     for (it = component->begin(); it != component->end(); ++it)
       (*it).Print();
     }
  END
  
  Note that the distance function is not predeclared.
  
  */


#include <stl.h>
#include "UnionList.h"
#include "../gamb++/MoleculeBase.h"

#include "Component.h"

template<class elementT> class ConnectedComponents{
public:
  //main function - see example on the top of the file for typical usage.
  UnionList< int, Component<elementT> > *FindCC(MoleculeBase& model,
                                            MoleculeBase& scene,
                                            vector<class elementT>& elements,
                                            float MaxDist);

  //calculate the distance between the two pairs.
  //by default the avrage between the distances on each molecule 
  virtual float Dist(elementT& i, elementT& j, 
                     MoleculeBase& model, MoleculeBase& scene)=0;
};



template <class elementT>
UnionList< int, Component<elementT> > *ConnectedComponents<elementT>::
FindCC(MoleculeBase& model, MoleculeBase& scene,
       vector<class elementT>& elements, float MaxDist){
  UnionList<int, Component<elementT> > *Components;
  Components=new UnionList<int, Component<elementT> >;
  
  //create unions - this phase is nessecary because I use Component::join
  //before I use the unite method so the unions are assumed to exist at 
  //that stage.
  for (int i=0; i < (int)elements.size(); ++i){
    elementT& e=elements[i];
    Component<elementT> *c = new Component<elementT>(e);
    Components->addUnion(i, *c);
  }
  
  UnionList<int, Component<elementT> >::iterator i_it, j_it;
  
  for (int i=0;i < (int)elements.size(); ++i){
    for (int j=i+1;j < (int)elements.size(); ++j){
      if (Dist(elements[i], elements[j], model, scene) < MaxDist){
        i_it=Components->getUnion(i);
        j_it=Components->getUnion(j);
        if (i_it != j_it){
          (*i_it).join(*j_it);
          Components->unite(i, j);
        }
      }
    }
  } 
  
  return Components;
}

#endif




