/* $Id$ */

#ifndef _Molecule_h

#define _Molecule_h

#include <vector.h>
#include "PDB.h"
#include "RigidTrans3.h"
#include "MoleculeBase.h"

template<class ParticleT>
/*
CLASS
  Molecule

  Template defines a collection of Particles that make up a Molecule.
  Among others class hosts a PDB file readr and linear algebra operators.

KEYWORDS
  particle, molecule, atom, vector, linear, rigid, matrix, centroid, PDB,
  file, rotation, translation

AUTHORS
  Meir Fuchs. (meirfux@math.tau.ac.il) and Zipi Fligelman (zipo@math.tau.ac.il)

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1997-9.

CHANGES LOG
<UL>
<LI>
</UL>

GOALS
  Serves as a container to a set of similarly typed class Particle descendants.
  The Molecule container hosts all methods of the STL vector container. It
  allows linear transfomrations to be applied to all Particles in the molecule.
  A PDB reader, enabling selected atoms to be read off a PDB format file is
  also supplied.

USAGE
  Construction is done using the default constructor. An empty Molecule object
  is returned.

  The Molecule may be built using the add method or using any other method
  supplied by the STL vector. Alternatively Particles may be added from a PDB
  format file using the readPDBfile method. This method may recieve a
  function object which will determine which of the atoms are loaded into
  memory.

  Matrices, rigid transformations, vector additions may be applied to all
  particles simultaneously. See the operators.

  The following is an example of using a Molecule<Particle> object to read in
  all C-alpha atom from a PDB file and print their coordinates after a
  rigid-transformation is applied to all the particles.
  EXAMPLE
    class CAlphaSelector : public PDB::Selector
    {  // This class defines a selector that will pick only C-alpha atoms.
    public:
      bool operator()(const char *PDBrec) const {
        const char *type = PDB::atomType(PDBrec);
        return (*(type+1) == 'C' && *(type+2) == 'A');
      }
    };

    main() {
      Molecule<Particle> M;
      ifstream pdb("pdb1tpo.ent");  
      M.readPDBfile(pdb, CAlphaSelector());
      pdb.close();
      RigidTrans3 rt(Whatever);
      M*= rt;
      for (Molecule<Particle>::iterator i= M.begin(); i != M.end(); i++)
        cout << *i << '\n';
    }
  END
  Note that Molecule has 2 base classes. The MolculeBase class is a virtual
  class that has a sole virtual method - the () operator. This class may be
  used to pass Molecule<T> objects of unknown type T into functions that
  access Molecule<T> vector positions only. This tool is used in the Match
  class.
*/
class Molecule : 
   public vector<ParticleT> , 
   public MoleculeBase
{
public:
    // GROUP: Constructors
    
    //// Default constructor - empty molecule.
    Molecule() : vector<ParticleT>() {}
    
    virtual ~Molecule() {}
    // GROUP: Adding Particles.
    
    //// readPDBfile accepts an input stream (file or cin) and an atom selector.
    // The default atom selector defined under PDB - PDB::Selector selects all
    // atoms by default. To write different selection functions one has to
    // define descendants to PDB::Selector() and redefine the operator()(const
    // char* PDBrec) const function to operate as desired.
    int 
    readPDBfile(istream& PDBstream, 
                const PDB::Selector& selector = PDB::Selector());
    
    //// Shorthand for vector push_back. Adds a new Particle at the end of the
    // Particle array.
    void add(const ParticleT& p);
    
    // GROUP: Inspection.
    
    //// The () operator returns a Particle's position according to index. 
    // The virtual operator() is declared in MoleculeBase
    // a virtual class that serves as a base for Molecule. The operator() is
    // used by functions recieving Molecule objects of unknown types. See for
    // example Match.h.
    Vector3 operator()(unsigned int particle) const;
    
    //// Returns centroid of all particle positions.
    Vector3 centroid() const;
    
  // GROUP: Operators.
    
    //// Transforms molecule using a linear transformation of type Matrix3.
    void transform(const Matrix3 &lt);
    
    //// Moves all atom coordinates by given vector3 move.
    inline void translate(const Vector3 &move);
    
    //// Applies a rigid transformation on all molecule's Particles.
    void rigidTrans(const RigidTrans3 &rt);
    
    //// Calculates the principal components of the molecules particles.
    // Returns the axis system along which the covariances are 0. Also
    // returns the variance along each axis. 
    void getPrincipalComponents(Matrix3& axes, Vector3& variance);

    //// Concating two molecules together molecule m is added at the
    // end of *this molecule
    void concat(const Molecule& m);
    
    //// Splitting a molecule on index i all the particles
    // from 0-i (included) will stay in this molecule
    // the other particles : i+1-end-of-molecule are returned
    // as a different molecule 
    Molecule splitOnIndex(const int i);
    
    
    //// Moves all atom coordinates by given vector3 move.
    Molecule& operator+=(const Vector3 &v);
    
    //// Applies a linear transformation on all protein coordinates.
    Molecule& operator*=(const Matrix3 &lt);
    
    //// Applies a rigid transformation on all atom coordinates.
    Molecule& operator*=(const RigidTrans3 &rt);
    
    //// Equality operator.
    Molecule& operator=(const Molecule& m);
    
    //// Output operator assuming the ParticleT class has its own
    // overload of the output operator.
    friend ostream& operator<<<>(ostream& s, const Molecule& m);
};


template<class ParticleT>
int Molecule<ParticleT>::readPDBfile(istream &PDBstream, 
				     const PDB::Selector& selector)
{
  char line[85];
  while (!PDBstream.eof()) {
    PDBstream.get(line, 85, '\n');
    // check that line is an ATOM rec and that selector accepts line.
    // if this is the case construct a new Particle using line and add the
    // Particle to the Molecule.
    if (PDB::isATOMrec(line) && selector(line)) 
      add(ParticleT(line));
    PDBstream.get(*line);  // Clear new-line from PDBstream.
  }
  return size();
}  

template<class ParticleT>
void Molecule<ParticleT>::add(const ParticleT& p)
{
  push_back(p);
}

template<class ParticleT>
Vector3 Molecule<ParticleT>::operator()(unsigned int particle) const 
{
  return (*(begin()+particle)).position();
}

template<class ParticleT>
Vector3 Molecule<ParticleT>::centroid() const 
{
  Vector3 sum;
  for (const_iterator i = begin(); i!=end(); ++i)
    sum+= *i;
  sum/=size();
  return sum;
}

template<class ParticleT>
void Molecule<ParticleT>::transform(const Matrix3& lt) 
{
  for (iterator i=begin(); i!=end(); ++i)   (*i)*=lt;
}

template<class ParticleT>
void Molecule<ParticleT>::translate(const Vector3 &move) 
{
  for (iterator i=begin(); i!=end(); ++i)   (*i)+=move;
}

template<class ParticleT>
void Molecule<ParticleT>::rigidTrans(const RigidTrans3& rt) 
{
  for (iterator i=begin(); i!=end(); ++i)   (*i)*=rt;
}
 
template<class ParticleT>
void Molecule<ParticleT>::getPrincipalComponents(Matrix3& axes, 
                                                 Vector3& variance)
{
  Vector3 mean = centroid();
  Matrix3 cov;
  for(const_iterator i= begin(); i!= end(); ++i) {
    Vector3 v = (*i).position() - mean;
    cov+= Matrix3(v, v);
  }
  cov/= size();
  Matrix3 dummy;
  cov.svd(dummy, variance, axes);  
}

template<class ParticleT>
void Molecule<ParticleT>::concat(const Molecule& m)
{
    for (iterator i = const_cast<ParticleT *>(m.begin()); i!= m.end(); ++i)
        push_back(*i);
}

template<class ParticleT>
Molecule<ParticleT> Molecule<ParticleT>::splitOnIndex(const int i)
{
    if (i > size())
        return Molecule();
    Molecule mol;
    for (iterator it = begin()+i; it != end(); ++it)
        mol.push_back(*it);
    erase(begin()+i,end());
    return mol;
}

template<class ParticleT>
Molecule<ParticleT>& Molecule<ParticleT>::operator+=(const Vector3 &v)  
{ 
  translate(v); 
  return *this;
}

template<class ParticleT>
Molecule<ParticleT>& Molecule<ParticleT>::operator*=(const Matrix3 &lt) 
{ 
  transform(lt);
  return *this;
} 

template<class ParticleT>
Molecule<ParticleT>& Molecule<ParticleT>::operator*=(const RigidTrans3 &rt)  
{ 
  rigidTrans(rt); 
  return *this;
}

template<class ParticleT>
Molecule<ParticleT>& Molecule<ParticleT>::operator=(const
                                                    Molecule<ParticleT>& m) {
  return (Molecule<ParticleT>&)vector<ParticleT>::operator=(m);
}

template<class ParticleT>
ostream& operator<<(ostream& s, const Molecule<ParticleT>& m) 
{
  for (vector<ParticleT>::iterator it = const_cast<ParticleT *>(m.begin());
       it != m.end() ; ++it) {
    s << *it << endl;
  }
  return s;
}
#endif














