#include <algobase.h>
#include "Particle.h"
#include "SurfacePoint.h"
#include "Shou.h"

static const char rcsid[]="$Id$";

typedef SurfacePoint::SurfaceType SurfaceType;

void sortAtoms(short atoms[3])
{
  if (atoms[1] > atoms[0]) swap(atoms[0], atoms[1]);
  if (atoms[2] > atoms[1]) swap(atoms[1], atoms[2]);
  if (atoms[1] > atoms[0]) swap(atoms[0], atoms[1]);
}  

SurfacePoint::SurfacePoint(): Particle(), norm()
{
  atoms[0] = atoms[1] = atoms[2] = -1;
}

SurfacePoint::SurfacePoint(const SurfacePoint& t)
  : Particle(t), norm(t.norm), area(t.area)
{
  atoms[0] = t.atoms[0];
  atoms[1] = t.atoms[1];
  atoms[2] = t.atoms[2];
}

SurfacePoint::SurfacePoint(const Vector3& newPos, const Vector3& newNorm, 
			   const float surface = 0.0,
                           const short atomIndex1 = -1, 
                           const short atomIndex2 = -1, 
                           const short atomIndex3 = -1) 
  : Particle(newPos), norm(newNorm), area(surface)
{
  setAtomIndices(atomIndex1, atomIndex2, atomIndex3);
}

SurfacePoint::SurfacePoint(const char* const shouRec)
  : Particle(Vector3(Shou::xCoord(shouRec), Shou::yCoord(shouRec), 
		     Shou::zCoord(shouRec))),
    norm(Vector3(Shou::xNorm(shouRec), Shou::yNorm(shouRec), 
		 Shou::zNorm(shouRec))),
    area(Shou::surface(shouRec))
{
  for (int i=0; i<3 ; ++i)
    atoms[i] = Shou::atomIndex(shouRec, i)-1; 
  sortAtoms(atoms);
}

Vector3 SurfacePoint::normal() const 
{
  return norm;
}

float SurfacePoint::surfaceArea() const
{
  return area;
}

void SurfacePoint::setAtomIndices(const short atomIndex1 = -1, 
                                  const short atomIndex2 = -1, 
                                  const short atomIndex3 = -1)
{
  atoms[0] = atomIndex1;
  atoms[1] = atomIndex2;
  atoms[2] = atomIndex3;
  sortAtoms(atoms);
}

short SurfacePoint::atomIndex(const unsigned short indx) const
{
  return atoms[indx];
}

bool SurfacePoint::atomIndicesGTE(const unsigned short atomIndex) const
{
  bool result = (atoms[0] >= atomIndex);
  if (result && (atoms[1] >= 0)) {
    result = (atoms[1] >= atomIndex);
    if (result && (atoms[2] >= 0))
      result = (atoms[2] >= atomIndex);
  }   
  return result;
}

bool SurfacePoint::atomIndicesLT(const unsigned short atomIndex) const
{
  return ((atoms[0] < atomIndex) && (atoms[1] < atomIndex) && 
          (atoms[2] < atomIndex));
}

SurfaceType SurfacePoint::surfaceType() const
{
  if (atoms[0] == -1)
    return Undef;
  if (atoms[1] == -1) 
    return Caps;
  if (atoms[2] == -1)
    return Belts;
  return Pits;
}

void SurfacePoint::orient(const Vector3 &v)
{
  norm= v/v.norm();
}

bool SurfacePoint::adjacentTo(const SurfacePoint& sp) const
{
  bool result = false;
  switch (surfaceType()) {
  case Caps:
    result =  (atoms[0] == sp.atoms[0] ||
               atoms[0] == sp.atoms[1] ||
               atoms[0] == sp.atoms[2]);
    break;
  case Belts:
    switch (sp.surfaceType()) {
    case Caps:
      result = (atoms[0] == sp.atoms[0] || atoms[1] == sp.atoms[0]);
      break;
    case Belts:
      result = (atoms[0] == sp.atoms[0] && atoms[1] == sp.atoms[1]);
      break;
    case Pits:
      result = ((atoms[0] == sp.atoms[0] && atoms[1] == sp.atoms[1]) ||
                (atoms[0] == sp.atoms[1] && atoms[1] == sp.atoms[2]) ||
                (atoms[0] == sp.atoms[0] && atoms[1] == sp.atoms[2]));
      break;
    case Undef:
      break;
    }
    break;
  case Pits:
    switch (sp.surfaceType()) {
    case Caps:
      result = (atoms[0] == sp.atoms[0] ||
                atoms[1] == sp.atoms[0] ||
                atoms[2] == sp.atoms[0]);
      break;
    case Belts:
      result = ((atoms[0] == sp.atoms[0] && atoms[1] == sp.atoms[1]) ||
                (atoms[1] == sp.atoms[0] && atoms[2] == sp.atoms[1]) ||
                (atoms[0] == sp.atoms[0] && atoms[2] == sp.atoms[1]));
      break;
    case Pits:
      result = (atoms[0] == sp.atoms[0] &&
                atoms[1] == sp.atoms[1] &&
                atoms[2] == sp.atoms[2]);
      break;
    case Undef:
      break;
    }
  case Undef:
    break;
  }
  return result;
}

SurfacePoint& SurfacePoint::operator*=(const Matrix3& lt) { 
  moveto(lt*position());
  orient(lt*normal());
  return *this;
}

SurfacePoint& SurfacePoint::operator*=(const RigidTrans3& rt) { 
  moveto(rt*position());
  orient(rt.rotation()*normal());
  return *this;
}


ostream& operator<<(ostream& s, const SurfacePoint& sp)
{
  s.setf(ios::fixed, ios::floatfield);
  s.setf(ios::right, ios::adjustfield);
  s.precision(3);
  
#ifdef __Debug
  s << "address=" << &sp << endl;
  s << "at0  at1  at2  pos.x   pos.y   pos.z   area    norm.x  norm.y  norm.z" 
    << endl;
#endif

  s.width(5);
  s << sp.atoms[0]+1; 
  s.width(5);
  s << sp.atoms[1]+1;
  s.width(5);
  s << sp.atoms[2]+1; 
  s.width(8);
  s << sp.pos.x(); 
  s.width(8);
  s << sp.pos.y();
  s.width(8);
  s << sp.pos.z();
  s.width(8); 
  s << sp.area;
  s.width(7);
  s << sp.norm.x();
  s.width(7);
  s << sp.norm.y();  
  s.width(7); 
  s << sp.norm.z();
  s << endl;
  return s;
}






