/* $Id: pdb.h,v 1.1 1997/06/02 16:19:45 meirfux Exp $ */

#ifndef _PDB_h

#define _PDB_h

/*
CLASS
  PDB

  The PDB class defines a set of tools used for reading the Protein Data Bank
  file format, used to describe protein structures.

KEYWORDS
  PDB, protein, record, atom, residue, structure, coordinate, field

AUTHORS
  Meir Fuchs. (meirfux@math.tau.ac.il)

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1997.

CHANGES LOG
<UL>
<LI>
</UL>

GOALS
  Class PDB is designed to encapsulate a set of methods used for reading the
  PDB file format. Methods are static requiring no object instantiation
  before their use.

USAGE
  A PDB file is comprised of 80-char lines called records. These records
  describe different properties of the protein. A record can describe an
  atom coordinate, an amino-acid sequence, an alpha helix or some other
  chemical or geometrical properties of the protein, depending on the
  record's type. For more information see the PDB_site.

  Class PDB's methods are all static. Hence the class's constructor was
  declared private so that no objects of type PDB will be declared. PDB's
  methods recieve a PDB record line as an argument and return requested data
  found in the record. 
  EXAMPLE    
    char rec[81];
    pdbFile.get(line, 81, '\n');         // read line from stream.
    if (PDB::isATOMrec(rec))             // if record is of type ATOM
      Vector3 v(PDB::atomXCoord(rec),    // instantiate a vector with
                PDB::atomYCoord(rec),    // atom's coordinates.
                PDB::atomZCoord(rec));
  END
  Note that one must make sure the record is of the correct type before asking
  for data found in the record. For example you cannot ask for an x-coordinate
  of an atom before making sure that you are dealing with an ATOM record.
  Method names are prefixed by the type of record they are meant to handle
  (e.g. atomXCoord is meant to extract the X coordinate from an ATOM record).
  Methods beginning with the is prefix (e.g. isATOMrec) check for record type.
*/
class PDB {

public:
  // GROUP: PDB record type determination methods.
 
  //// Returns true is given record is an ATOM record.
  static bool isATOMrec(const char *const PDBrec);

  //// Returns true is given record is a HETATM record.
  static bool isHETATMrec(const char *const PDBrec);

  //// Returns true is given record is an ATOM record of a C-alpha atom.
  static bool isCaATOMrec(const char *const PDBrec);


  // GROUP: ATOM record costants and methods.

  static const unsigned short atomIndexField    =   6;
  static const unsigned short atomTypeField     =  12;
  static const unsigned short atomResTypeField  =  17;
  static const unsigned short atomChainIdField  =  21;
  static const unsigned short atomResIndexField =  22;
  static const unsigned short atomXCoordField   =  30;
  static const unsigned short atomYCoordField   =  38;
  static const unsigned short atomZCoordField   =  46;

  //// Returns an ATOM record's X coordinate.
  static float atomXCoord(const char *const PDBrec);

  //// Returns an ATOM record's Y coordinate.
  static float atomYCoord(const char *const PDBrec);

  //// Returns an ATOM record's Z coordinate.
  static float atomZCoord(const char *const PDBrec);

  //// Returns an ATOM record's atom index number.
  static short atomIndex(const char *const PDBrec);

  //// Returns a pointer to an ATOM's record atom type string.
  // The atom type is a 4 charachter long field. A pointer is returned to
  // the correct position withing the given string PDBrec. Hence the pointer
  // should not be freed or deleted.
  static const char* atomType(const char *const PDBrec);
  // 4 chars

  //// Returns an ATOM record's chain id.
  static char  atomChainId(const char *const PDBrec);

  //// Returns an ATOM record's residue index.
  static short atomResidueIndex(const char *const PDBrec);

  //// Returns an ATOM record's residue type in single letter code (As opposed
  // to 3 letter code).
  static char  atomResidueType(const char *const PDBrec);

  //// Given a residue char ID returns a 3-letter residue mnemonic.
  static const char* const residueLongName(const char resChar);

  //// Selector is a general purpose class used to select records from a PDB
  // file. Using descendants of this class one may implement arbitrary
  // selection functions with operator() and pass them to PDB reading functions
  // for object selection. (See Molecule for examples).
  class Selector {
  public:
    virtual bool operator()(const char*) const {  return true;  }
  };

private:
  // Private constructor to prevent object instantiation. Class was meant to
  // be use as a static method pool.
  PDB() {}

  // Given a charachter string with a PDB residue type code (3 letters)
  // residueShortName will return the single letter residue code.
  // residueShortName uses the following constants in order to seach for the
  // appropriate residue name. If no matching resiude is found the function
  // will return 'X'.
  static char residueShortName(const char *const resName);

};

#endif 

