#include "iostream.h"
#include "strstream.h"
#include "vector.h"
#include "Parameters.h"

typedef Parameters::ParamType ParamType;

class Parameter
{
public:
    
    Parameter();
    Parameter(const string& name, const int val);
    Parameter(const string& name, const float val);
    Parameter(const string& name, const string& val);
    Parameter(const Parameter& param);

    ~Parameter() {}
    const string& paramLabel() const;
    ParamType paramType() const;
    
    int intValue() const;
    float floatValue() const;
    const string& stringValue() const;
    
    Parameter& operator=(const Parameter& param);
    
    friend bool operator<(const Parameter& p, const Parameter& q);
    friend ostream& operator<<(ostream& out, const Parameter& p);
    
private:
    string paramName;
    ParamType type;
    union {
        int i;
        float f;
    };
    string s;
};

typedef vector<Parameter> ParamList;

static ParamList paramList;

Parameter::Parameter() : paramName(), type(Parameters::Undef)
{}

Parameter::Parameter(const string& name, const int val) 
{
  paramName = name;
  i = val;
  type = Parameters::Int;
}

Parameter::Parameter(const string& name, const float val) 
{
    paramName = name;
    f = val;
    type = Parameters::Float;
}

Parameter::Parameter(const string& name, const string& val) 
{
  paramName = name;
  s = val;
  type = Parameters::Str;
}
Parameter::Parameter(const Parameter& param)
{
    paramName = param.paramName;
    type = param.type;
    if (type == Parameters::Int)
        i = param.i;
    else
        f = param.f;
    s = param.s;
}


const string& Parameter::paramLabel() const
{
    return paramName;
}

ParamType Parameter::paramType() const
{
  return type;
}

int Parameter::intValue() const 
{
  return i;
}

float Parameter::floatValue() const 
{
  return f;
}

const string& Parameter::stringValue() const 
{
  return s;
}

Parameter& Parameter::operator=(const Parameter& param)
{
    paramName = param.paramName;
    type = param.type;
    if (type == Parameters::Int)
        i = param.i;
    else
        f = param.f;
    s = param.s;
    return *this;
}

bool operator<(const Parameter& p, const Parameter& q)
{
  return (p.paramName < q.paramName);
}
  
ostream& operator<<(ostream& out, const Parameter& p) {
  switch(p.type) {
  case Parameters::Int:
    return out << p.paramName << '\t' << "(Int)" << '\t' << p.i;
  case Parameters::Float:
    return out << p.paramName << '\t' << "(Float)" << '\t' << p.f;
  case Parameters::Str:
    return out << p.paramName << '\t' << "(Str)" << '\t' << p.s;
  case Parameters::Undef:
    break;
  }
  return out << '\n';
}
    

ParamList::iterator findInsertionPoint(const ParamList& paramList,
                                       const string& paramName)
{
  unsigned short start = 0;
  unsigned short stop = paramList.size();
  while (stop != start) {
    unsigned short pos = start + (stop-start)/2;
    int comp = paramName.compare(paramList[pos].paramLabel());
    if (comp == 0)
      stop = start = pos;
    else if (comp > 0)
      start = pos + 1;
    else
      stop = pos;
  }
  return (ParamList::iterator)(paramList.begin()+stop);
}

Parameters::Parameters()
{}
  
void Parameters::readParameters(istream& paramStream)
{
  char line[256];
  char c;

  while (!paramStream.eof()) {
    paramStream.get((char*)line, 256, '\n');
    paramStream.get(c);
    string param(line);
    string paramName = nextToken(param);
    if (paramName.length() == 0) continue;
    if (*(paramName.data()) == '#') continue;
    updateParameter(paramName, param.c_str());
  }  
}

void Parameters::addParameter(const string& paramName, const int value)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if (pos != paramList.end() && (*pos).paramLabel() == paramName)
    (*pos) = Parameter(paramName, value);
  else
    paramList.insert(pos, Parameter(paramName, value));
}

void Parameters::addParameter(const string& paramName, const double value)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if (pos != paramList.end() && (*pos).paramLabel() == paramName)
    (*pos) = Parameter(paramName, (float)value);
  else
    paramList.insert(pos, Parameter(paramName, (float)value));
}

void Parameters::addParameter(const string& paramName, const string& value)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if (pos != paramList.end() && (*pos).paramLabel() == paramName)
    (*pos) = Parameter(paramName, value);
  else
    paramList.insert(pos, Parameter(paramName, value));
}

void Parameters::updateParameter(const string& paramName, 
                                 const char* const value)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if (pos != paramList.end() && (*pos).paramLabel() == paramName)
    switch ((*pos).paramType()) {
    case Int:
      (*pos) = Parameter(paramName, atoi(value));
      break;
    case Float:
      (*pos) = Parameter(paramName, (float)atof(value));
      break;
    case Str:
      (*pos) = Parameter(paramName, string(value));
    case Undef:
      (*pos) = Parameter(paramName, string(value));
    }
  else
    paramList.insert(pos, Parameter(paramName, string(value)));
}
  

ParamType Parameters::paramType(const string& paramName)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if ((*pos).paramLabel() == paramName)
    return (*pos).paramType();
  else
    return Undef;
}
  

int Parameters::getInt(const string& paramName)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if ((*pos).paramLabel() == paramName)
    switch ((*pos).paramType()) {
    case Int:
      return (*pos).intValue();
    case Float:
      return (int)(*pos).floatValue();
    case Str:
      return atoi((*pos).stringValue().data());
    case Undef:
      break;
    }
  return 0;
}
  
float Parameters::getFloat(const string& paramName)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if ((*pos).paramLabel() == paramName)
    switch ((*pos).paramType()) {
    case Float:
      return (*pos).floatValue();
    case Int:
      return (float)(*pos).intValue();
    case Str:
      return atof((*pos).stringValue().data());
    case Undef:
      break;
    }
  return 0.0;
}

string Parameters::getString(const string& paramName)
{
  ParamList::iterator pos = findInsertionPoint(paramList, paramName);
  if ((*pos).paramLabel() == paramName)
    switch ((*pos).paramType()) {
    case Str:
      return (*pos).stringValue();
    case Float: {
      char num[30];
      ostrstream f2str(num, 30);
      f2str << (*pos).floatValue() << '\0';
      return string(num); 
    }
    case Int: {
      char num[30];
      ostrstream i2str(num, 30);
      i2str << (*pos).intValue() << '\0';
      return string(num); 
    }
    case Undef:
      break;
    }
  return "";
}

void Parameters::dump(ostream& out) 
{
  for (ParamList::iterator i=paramList.begin(); i != paramList.end(); ++i)
    out << *i << '\n';
}

void Parameters::dump(DebugStream& out, const unsigned int msgLevel) 
{
  for (ParamList::iterator i=paramList.begin(); i != paramList.end(); ++i)
    out(msgLevel) << *i;
}

string Parameters::nextToken(string& str)
{
  unsigned int start = 0;
  while (start < str.length() && 
         (str[start] == ' ' || str[start] == '\t' || str[start] == '\n'))
    ++start;
  
  if (start >= str.length()) {
    str = "";
    return "";
  }
  
  unsigned int stop = start+1;
  while (stop < str.length() && 
         str[stop] != ' ' && str[stop] != '\t' && str[stop] != '\n')
    ++stop;

  unsigned int next = stop;
  while (next < str.length() && 
         (str[next] == ' ' || str[next] == '\t' || str[next] == '\n'))
    ++next;

  string result = str.substr((int)start, stop-start);
  str = str.substr((int)next);
  return result;
}
