#include "RMSDCluster.h"

typedef RMSDCluster::Transformation Transformation;

RigidTrans3 Transformation :: trans_parameter_average(RigidTrans3 *T1,
                                                      RigidTrans3 *T2, 
                                                      float p, float q) {
    { //normalize weights to be the relative weight in the outcome.
        float wsum=p+q;
        p /= wsum;
        q /= wsum;
    }
    return RigidTrans3(T1->rotationAngles() * p + T2->rotationAngles() * q,
                       T1->translation() * p + T2->translation() * q);
}

//create a transformation which is the avrage.
/*
 * function trans_average:
 * takes 2 (close) rigid transformations and make a "weighted average" 
 * between them. We take as "weighted average" the result of the 
 * transformation. but since we need a rigid (orthogonal..) transformation 
 * a simple average is not enough.
 * our avarage promises these properties: 
 *   1) the projection of the  X_Y plane is the exact avarage (of T1 and 
 *      T2 projections)
 *   2) the X and Y projection are as close to the avarage (of the original 
 *      projections) as can be.
 *   3) the Z vector is at an (dX cross dy) offset from the avarage of the 
 *      diferrent Z zector projections.
 * the algorithm is this:
 * 1) find the new X_Y plane ,from { qT1(e_x)+pT2(e_x), qT1(e_y)+pT2(e_y) }.
 * 2) in it, take the sum of these vectors as a first diagonal (the line X=Y) 
 *    this is done to maintain symetry between the X and the Y axis.
 * 3) compute it's orhtogonal, and with it the new X and Y projections.
 * 
 */ 
RigidTrans3 Transformation :: trans_average(RigidTrans3 *T1,
                                            RigidTrans3 *T2, 
                                            float p, float q) {
    { //normalize weights to be the relative weight in the outcome.
        float wsum=p+q;
        p /= wsum;
        q /= wsum;
    }
    
    static Vector3 e_x(1.0, 0.0, 0.0);
    static Vector3 e_y(0.0, 1.0, 0.0);
    static Vector3 e_z(0.0, 0.0, 1.0);
    static Vector3 zero(0.0, 0.0, 0.0);
    
    Vector3 t1_o = (*T1)*zero;
    Vector3 t1_x = (*T1)*e_x;
    Vector3 t1_y = (*T1)*e_y;
    t1_x -= t1_o;  t1_y -= t1_o; // want the offset from T1(O)
    
    Vector3 t2_o = (*T2)*zero;
    Vector3 t2_x = (*T2)*e_x;
    Vector3 t2_y = (*T2)*e_y;
    t2_x -= t1_o;  t2_y -= t1_o;
    
    Vector3 avg_tx = (q*t1_x) + (p*t2_x);
    Vector3 avg_ty = (q*t1_y) + (p*t2_y);
    Vector3 avg_to = (q*t1_o) + (p*t2_o); // this is the new Translation.
    
    Vector3 tz = avg_tx & avg_ty;  // new Z (e3) projection
    tz /= tz.norm();
    
    // half the main diagonal vector: (1/2 , 1/2 ,0)   
    Vector3 new_main_diag = (avg_tx+avg_ty);
    new_main_diag /= (new_main_diag.norm() * sqrt_2);// need only diag*Sqrt2 ...
    
    // half the secondary diagonal vector : (1/2 , -1/2 , 0)
    // this is exactly the vector product of new_diag and the 
    // Normal to the Plane (tz)
    Vector3 new_secondary_diag = new_main_diag & tz; 
    
    Vector3 tx =  new_main_diag + new_secondary_diag;
    Vector3 ty =  new_main_diag - new_secondary_diag;
    
    //create and return the new avarage transformation
    Matrix3 M(tx, ty, tz);
    M.transpose();          // tx ty and tz  are coloumn vectors in M
    
#ifdef _DEBUG_NOTHING
    cout << "Trans 1 : " << *T1 << " p=" << p << endl
         << "Trans 2 : " << *T2 << " q=" << q << endl
         << "Result Trans  : " << RigidTrans3(M,avg_to) << endl;
    cout << " det " << M.determinant() << endl;
#endif
    
    return RigidTrans3(M,avg_to); 
}

//join the two transformations using an average transformation with the sum
//of the scores.
void Transformation::join(Transformation& t){
    //cerr << "%";
    //rt = trans_average(&rt, &(t.rt), scr, t.scr);
    rt = trans_parameter_average(&rt, &(t.rt), scr, t.scr);
    scr += t.scr;
    joins++;
}

void RMSDCluster::PreProcess(vector<Particle>& Set){
    int n = Set.size();
    //set the Xmean;
    cerr << "inside preprocess" << endl;
    for (unsigned int k=0; k < Set.size(); k++){
        const Vector3& Xk = Set[k].position();
        Xmean += Xk;
        Xmean2 += Vector3(Xk.x() * Xk.x(), 
                          Xk.y() * Xk.y(),
                          Xk.z() * Xk.z());
        /*
          for (unsigned int i=0; i < 3; i++){
          for (unsigned int j=0; j < 3; j++) X[i][j] += Xk[i] * Xk[j];
          }
        */
    }

    Xmean /= n;
    Xmean2 /= n;

    cerr << "Xmean and Xmean2 are set" << endl;
    /*
      for (int i=0; i < 3; i++){
      for (int j=0; j < 3; j++){
      X[i][j] /= n;
      }
      }
    */

    cerr << "leaving preprocess" << endl;
}
/*
  //the original dist function as it is written in FLEXX paper.
  float RMSDCluster::Dist(RigidTrans3& t1, RigidTrans3& t2){
  Matrix3 A = t1.rotation();
  A.transpose();
  A*=t2.rotation();
  float res=0.0;
  
  for (int i=0; i < 3; i++){
  for (int j=0; j < 3; j++){
  res += (((i==j)? 1.0 : 0.0) - A[i][j] - A[j][i]) * X[i][j];
  }
  }
  
  res *= 2.0;
  
  Vector3 d = t1.translation() - t2.translation();
  res += d * d;
  
  Matrix3 B = t1.rotation();
  B -= t2.rotation();
  res += 2 * d * (B * Xmean);
  
  return sqrt(res);
  }
*/

void RMSDCluster::squareVector(Vector3& v){
    multby(v, v);
}

void RMSDCluster::multby(Vector3& v, Vector3& u){
    v.updateX(v.x() * u.x());
    v.updateY(v.y() * u.y());
    v.updateZ(v.z() * u.z());
}

float RMSDCluster::Dist(RigidTrans3& t1, RigidTrans3& t2){
    Matrix3 B=t1.rotation();
    B -= t2.rotation();

    Vector3 b1=B.row(1);
    Vector3 b2=B.row(2);
    Vector3 b3=B.row(3);

    squareVector(b1);
    squareVector(b2);
    squareVector(b3);

    Vector3 d=t1.translation() - t2.translation();
    Vector3 d2 = d;
    squareVector(d2);
    Vector3 tmp = t1.rotation() * t1.translation() 
        + t2.rotation() * t2.translation()
        - t1.rotation() * t2.translation() 
        - t2.rotation() * t1.translation();
    multby(tmp, Xmean);
    tmp *= 2;

    Vector3 resvec(b1 * Xmean2, b2 * Xmean2, b3 * Xmean2);
    resvec += tmp;
    resvec += d2;
    
    float dist=sqrt(resvec.x() + resvec.y() + resvec.z());
    //cerr << "dist is : " << dist << " t1 " << t1 << " t2 " << t2 << endl;
    return dist;
}

/*
  One iteration of the clustering algorithm:
  create a graph of the transformations : each transformation is a vertex and the 
  weights on the edges are the distances between the transformations.
  Go through the edges in acending lengths and every contract edges that are less than
  MaxDist long, but only use each vertex once. (throughout the iteration).

  For fast creation of the graph create a hash and put the picture of the Xmean 
  by each transformation. Then calculate distance only between transformation which 
  the Xmean is less than 2*MaxDist.
*/

int RMSDCluster::fastWork(float MaxDist,
                          vector<Transformation *>& trans){

    static int iteration=0;
    iteration++;
    int joins=0;
    bool used[trans.size()];

    GeomHash<Vector3, int> gHash(3, BinSize);

    //load the hash
    vector<Transformation*>::iterator i;
    for (i = trans.begin(); i != trans.end(); ++i){
        used[i-trans.begin()] = false;
        gHash.insert((*i)->rt * Xmean, i-trans.begin());
    }

    cerr << "Hash created" << endl;

    typedef vector<int> Bucket;
    vector<Bucket *> *buckets = gHash.getBuckets();

    cerr << "no. of buckets is : " << buckets->size() << endl;
    
    int first_index=0;
    vector<Bucket *>::iterator b;
    for (b = buckets->begin(); b != buckets->end(); ++b){
        cerr << "bucket size is : " << (*b)->size() << endl;
        //edges vector is defined inside the loop so that it will be a new 
        //vector for every bin
        vector<Edge> E;
        //for each bin
        Bucket::iterator it, it_j;
        for (it = (*b)->begin(); it != (*b)->end(); ++it){
            for (it_j=it+1; it_j != (*b)->end(); ++it_j){
                float d = Dist(trans[*it]->rt, trans[*it_j]->rt);
                if (d < MaxDist){
                    E.push_back(Edge(*it, *it_j, d));
                }
            }
        }

        //cluster all transformations in the bin
        sort(E.begin(), E.end());
        
        vector<Edge>::iterator e;
        for (e=E.begin(); (e != E.end()); e++){
            //check if any end of the edge is deleted
            if (!used[e->t1] && !used[e->t2] && e->len < MaxDist){
                joins++;
                used[e->t1]=true;
                used[e->t2]=true;
                trans[e->t1]->join(*(trans[e->t2]));
                trans[e->t2]->valid=false;
            }
        }
        cerr << "#";  
        first_index += (*b)->size();
    }
    return joins;
}

int RMSDCluster::Work(float MaxDist,
                      vector<Transformation *>& trans)
{
    int joins=0;

    vector<Edge> E;
    bool used[trans.size()];

    GeomHash<Vector3, int> gHash(3, MaxDist);
    vector<Transformation*>::iterator i;
    for (i = trans.begin(); i != trans.end(); ++i){
        used[i-trans.begin()] = false;
        gHash.insert((*i)->rt * Xmean, i-trans.begin());
    }
   
    cerr << "Hash created" << endl;

    HashResult<int>::iterator it;
    for (i = trans.begin(); i != trans.end(); ++i){
        int index = i - trans.begin();
        HashResult<int> result;
        gHash.query((*i)->rt * Xmean, MaxDist*2, result);
        for (it = result.begin(); it != result.end(); ++it){
            //only create an edge if index < *it so that 
            //an edge will not be created twice
            if (index < (*it)){
                float d = Dist(trans[index]->rt, trans[(*it)]->rt);
                if (d < MaxDist) E.push_back(Edge(index, *it, d));
            }
        }
        cerr << "result size : " << result.size() 
             << " E.size() : " << E.size() << endl;
    }
    /*
      //build the graph
      vector<Transformation*>::iterator i,j;
      for (i=trans.begin(); i != trans.end(); ++i){
      //build edges
      for (j=i+1; j != trans.end(); ++j){
      float d=Dist((*i)->rt, (*j)->rt);
      if (d < MaxDist) E.push_back(Edge(i-trans.begin(), j-trans.begin(), d));
      }
      }
    */
    sort(E.begin(), E.end());

    vector<Edge>::iterator e;
    for (e=E.begin(); (e != E.end()); e++){
        //check if any end of the edge is deleted
        if (!used[e->t1] && !used[e->t2] && e->len < MaxDist){
            joins++;
            used[e->t1]=true;
            used[e->t2]=true;
            trans[e->t1]->join(*(trans[e->t2]));
            trans[e->t2]->valid=false;
        }
    }

    return joins;
}

vector<Transformation *> *RMSDCluster::Clean(vector<Transformation *> *trans){
    vector<Transformation *> *res = new vector<Transformation *>();
    for (unsigned int i=0; i < trans->size(); i++){
        if ((*trans)[i]->valid) res->push_back((*trans)[i]);
    }
    return res;
}
    
vector<RigidTrans3>& RMSDCluster::Cluster(float MaxDist, 
                                          vector<RigidTrans3>& trans){

    cerr << "inside Cluster - MaxDist is " << MaxDist << endl;
    vector<Transformation *> *T = new vector<Transformation *>();
    
    //create initial vectors of transfomation and bit vector to indicate what 
    //is deleted
    vector<RigidTrans3>::iterator it;
    for (it = trans.begin(); it != trans.end(); ++it){
        Transformation *t=new Transformation(*it, 1);
        T->push_back(t);
    }

    //the main loop
    if (BinSizeSet){
        //use fastWork first
        while (fastWork(MaxDist, *T)){
            cerr << ".";
            T=Clean(T);
        }
    }

    while (Work(MaxDist, *T)){
        cerr << ".";
        T=Clean(T);
    }
    

    //build the vector for output.
    vector<RigidTrans3> *res = new vector<RigidTrans3>();
    for (unsigned int i = 0; i < T->size(); ++i){
        res->push_back((*T)[i]->rt);
    }

    cerr << "leaving cluster" << endl;
    return *res;
}





