#ifndef __RMSDCLUSTER_H
#define __RMSDCLUSTER_H

#include "GeomHash.h"
#include "HashResult.h"
#include "RigidTrans3.h"
#include "Particle.h"
#include "Vector3.h"
#include <stl.h>

//#define LEDA_PREFIX
//#include <LEDA/graph_alg.h>

/*
CLASS
  RMSDCluster

  Class intended to provide a good quality clustering of transformation, 
  using the distance between two tranformation as the RMSD of the picture 
  on a set of points. (the set of points is given by the user for the 
  preprocessing).

  RMSDCluster provides two clustering algorithms - one is faster and less 
  accurate, while the other is more thorough but is limited by the number 
  of transformations it can handle.

  The decision of which algorithm to use is left to the user, who can determine 
  with the useFastClustering to use the faster algorithm. The user is also 
  required to supply the hash cube size for the fast clustering. The larger 
  the cube size the more accurate the algorithm is, but also there will be 
  more transformations in each bin and so the algorithm will be slower.

  Both algorithms are iterative tring to join pairs of close transformatins
  using a weighted average heuristic. The difference is that the thorough 
  algorithm calculates the entire distance matrix of all the transformations 
  at once and then continuessly contracts the smallest edge that connects 
  vertices that were not used in this iteration, while the faster algorithm 
  uses a hash table and at first only clusters transformations which put the 
  center of the molecule in close locations.

  The decision of which algorithm to use is left to the user, who can determine 
  with the useFastClustering

  Calculation of the distance between any two rigid transformation is O(1) in 
  complexity, after an O(N) preprocessing.

KEYWORDS
  cluster, RMSD, flexx, efficient, effective

AUTHORS
  Ram Nathaniel (ramn@math.tau.ac.il).

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1999.

CHANGES LOG
<UL>
<LI>
</UL>

GOALS
  The RMSDCluster class serves as an easy interface to a clustering paradigm 
  partially simmilar to the one introduced by FLEXX (after a few corrections 
  of the original furmulas).

  Complexity : 
    preprocessing O(N), (each distance calculation O(1)).
    Overall complexity is O(N^2 * log(N)) (but a small constant ...)

  There are actually two different clustering algorithms, one is thorough 
  and one is faster but less accurate.
  Both algorithms are iterative tring to join pairs of close transformatins
  using a weighted average heuristic. The difference is that the thorough 
  algorithm calculates the entire distance matrix of all the transformations 
  at once while the faster algorithm divides the transformations first by the 
  picture of the molecule center under the transformation and only then checks
  distances at each bin of separately. After the division is useless run the 
  first algorithm.

  The basic algorithm:
    while something was done
      create a graph from the transformation left in play.
      sort the edges (the distances).
      contract every edge smaller than MaxDist starting from the shortest
      BUT ONLY USE EACH VERTEX ONCE.

  The faster algorithm:
    calculate the picture of Xmean of the set under each transformation and 
      put in a GeomHash (the size of the cubeSize is given by the user)
    while something was done
      foreach bucket
        cluster the transformations in the bucket 
    call the basic algorithm.

  Notes:
    The distance between two transformation is calculated as the RMSD of the 
    pictures by the two transformations on a set of points supplied by the user. 

    The calculation of the average transformation is done as follows:
    1) calculate the plane formed by the weighted averages of X and Y axes.
    2) set the new X and Y axes so that the line X=Y is the exact weighted 
    average between the two transformations.
    3) set the Z axis in place.
    This should be accurate for close transformations and have no problems with 
    limits such as the parameter naive average does.

  This is possibly the most accurate clustering available given a correct choice 
  of base set by the user.

USAGE
  Usage of RMSDCluster is quite simple:
  First you have to define the referece set. This is done by invoking the method
  void PreProcess(vector<Particle>& set);

  Of course sending a Molecule<Particle> object will also do.
  Then invoke the Cluster method where the MaxDist parameter indicates the 
  maximal length of edge that will be contracted.

  EXAMPLE
    vector<RigidTrans3> trans;
    Molecule<Particle> mol;
  
    RMSDCluster clusterer();
    cluseterer.PreProcess(mol);
    vector<RigidTrans3 *> result=cluseterer.Cluster(trans, 5.5);
  END

  For use of the faster algorithm just add:
  clusterer.useFastClustering(cubeSize);
  before invoking the Cluster method.
*/

class RMSDCluster{
private:
    ////private class intended to avoid the need of the LEDA library in 
    //implementing the clustering phase.
    class Edge{
    public:
        int t1, t2;
        float len;
        
        Edge(){ t1=0; t2=0; len=0.0; }
        Edge(int first, int second, float l):
            t1(first), t2(second), len(l){}
        
        friend bool operator<(const Edge &e1, const Edge &e2){
            return (e1.len < e2.len);
        }
    };

    ////Private class intended to hold the transformations throughout the 
    //clustering process.
    class Transformation{
    public:
        //used to indicate deletions in function Work. Deletion will them be done
        //faster with the Clean method of RMSDCluster. Deletion in stl vectors is
        //very slow.
        bool valid; 

        ////score for the transformation - used in the weighted avrage, 
        //initialized to 1 by default.
        float scr;
        
        RigidTrans3 rt;

        //// The number of times this transformation and its ancessters were
        // involved in an edge contraction
        unsigned int joins;

        ////standard constructor.
        Transformation(RigidTrans3& rig, const float TransScore=1.0): 
            valid(true), scr(TransScore), 
            rt(rig), joins(0) {}

        ////avrages the tranfromations into this. Transformation t is unchanged.
        virtual void join(Transformation& t);

        inline unsigned int updates() const { return joins; }

    private:
        //// written by Guy Baruch & Ram Nathaniel (workshop 1998)
        // return an averaged transformation by the weights p and q.
        RigidTrans3 trans_average(RigidTrans3 *T1,
                                  RigidTrans3 *T2, 
                                  float p, float q);

        RigidTrans3 trans_parameter_average(RigidTrans3 *T1,
                                            RigidTrans3 *T2,
                                            float p, float q);
    };

public:
    ////default constructor
    RMSDCluster():BinSizeSet(false){};

    //// The main function which should be called after the preprocessing is done.
    vector<RigidTrans3>& Cluster(float MaxDist, vector<RigidTrans3>& trans);

    //// Preprocessing of a set of points in order to calculate distances on them.
    void PreProcess(vector<Particle>& Set);

    //// Returns the RMSD of the preprocessed set of points under the two 
    // transformations.
    virtual float Dist(RigidTrans3& t1, RigidTrans3& t2);

    //// Sets the bin size on which the clustering will be done at the first stage.
    // if not used the class will preform clustering on all bins at once.
    void useFastClustering(float binsize){ 
        BinSize = binsize; BinSizeSet = true; 
        cerr << "Bin size is set to " << binsize << endl;
    }

private:
    //square the coordinates of the vector v
    void squareVector(Vector3& v);

    //v becomes the vector whos coordinates are the numerical product of the 
    //coords of u and v. u is unchanged.
    inline void multby(Vector3& v, Vector3& u);

protected:

    //// The main loop builds the graph of transformations a contract the edges.
    virtual int Work(float MaxDist, vector<Transformation *>& trans);

    //// Just like Work only handles each bin independently until there is nothing
    // else to be done this way. Bins size is deremined by SetSpeed
    virtual int fastWork(float MaxDist, vector<Transformation *>& trans);

    //// Remove transformations which are not valid.
    // should be used after each invokation of Work.
    virtual vector<Transformation *> *Clean(vector<Transformation *> *trans);

private:
    bool BinSizeSet;
    float BinSize;
    Vector3 Xmean; //mean point of the set
    Vector3 Xmean2; //mean of points after squaring each coord.
    //float X[3][3]; //coefficiant
};

#endif






