#ifndef __RMSDCLUSTER_H
#define __RMSDCLUSTER_H

#include "GeomHash.h"
#include "HashResult.h"
#include "RigidTrans3.h"
#include "Particle.h"
#include "Vector3.h"
#include "Molecule.h"

#include <stl.h>

/*
CLASS
  RMSDCluster

  Template intended to provide a good quality clustering of
  transformations, using the distance between two tranformation as the
  RMSD of the picture on a set of points. (the set of points is given
  by the user in the preProcessing stage).

  RMSDCluster provides two clustering algorithms - one is faster and less 
  accurate, while the other is more thorough but is limited by the number 
  of transformations it can handle.
  
  The decision of which algorithm to use is left to the user, who can
  determine with the useFastClustering method to use the faster
  algorithm. The user is also required to supply the hash cube size
  for the fast clustering. The larger the cube size the more accurate
  the algorithm is, but also there will be more transformations in
  each bin and so the algorithm will be slower.

  Both algorithms are iterative tring to join pairs of close transformatins
  using a weighted average heuristic. The difference is that the thorough 
  algorithm calculates the entire distance matrix of all the transformations 
  at once and then continuessly contracts the smallest edge that connects 
  vertices that were not used in this iteration, while the faster algorithm 
  uses a hash table and at first only clusters transformations which put the 
  centroid of the molecule in close locations.

  Calculation of the distance between any two rigid transformation is
  O(n) in complexity, where n is the number of points, given in the
  pre-processing stage.

KEYWORDS
  cluster, RMSD

AUTHORS
  Ram Nathaniel   (ramn@math.tau.ac.il)
  Dina Duhovny    (duhovka@math.tau.ac.il)
  Oranit Shem-Tov (ornit@math.tau.ac.il)

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 2001

GOALS
  The RMSDCluster class serves as an easy interface to a clustering paradigm 
  partially similar to the one introduced by FLEXX

  Complexity : 
    preProcessing O(N) 
    Overall complexity is O(N*M^2 * log(M)) (but a small constant ...)
    where:
    - N is the number of particles, givan in the pre-processing stage
    - M is the number of transformations to be clustered

  There are actually two different clustering algorithms, one is thorough 
  and one is faster but less accurate.
  Both algorithms are iterative trying to join pairs of close transformatins
  using a weighted average heuristic. The difference is that the thorough 
  algorithm calculates the entire distance matrix of all the transformations 
  at once while the faster algorithm divides the transformations first by the 
  picture of the molecule centroid under the transformation and only then checks
  distances at each bin of separately. After the division is useless run the 
  first algorithm.

  The basic algorithm:
    while something was done
      create a graph from the transformation left in play.
      sort the edges (according to the distances).
      contract every edge smaller than maxDist starting from the shortest
      BUT ONLY USE EACH VERTEX ONCE.

  The faster algorithm:
    calculate the picture of centroid of the set under each transformation and 
      put in a GeomHash (the size of the cubeSize is given by the user)
    while something was done
      foreach bucket
        cluster the transformations in the bucket 
    call the basic algorithm.

  Notes:
    The distance between two transformation is calculated as the RMSD of the 
    pictures by the two transformations on a set of points supplied by the user. 

USAGE
  The RMSDCluster is a template which uses a class parameter known as TransT.
  TransT should be a class, which implements the following methods: 
  - void join(TransT& trans);
  - float getScore() const;
  - void incScore(float s); 
  - const RigidTrans3& rigidTrans();

  Usage of RMSDCluster is quite simple:

  First you have to define the reference set. This is done by invoking
  the method void preProcess(vector<Particle>& particles); Of course
  sending a Molecule<Particle> object will also do.

  Then invoke the Cluster method where the maxDist parameter indicates
  the maximal length of edge that will be constracted (i.e. the
  maxmimal distance allowed between two transformations in the same
  cluster)

  EXAMPLE

  class Transformation {
     private:
        RigidTrans3 rt;
        float score;
     public{	 
        Transformation() {...}

        void join(Transformation& t) {...}
	  
        float getScore() const {
	   return score;
	}
    
	void incScore(float s) const {
	   score += s;
	}

	const RigidTrans3& rigidTrans() const {
	   return rigidTrans;
	}
    };

    vector<Transformation> trans;
    Molecule<Particle> mol;
  
    RMSDCluster<Transformation> clusterer();
    cluseterer.preProcess(mol);
    vector<Transformation>& result = cluseterer.Cluster(trans, 5.5);
  END

  For use of the faster algorithm just add:
  clusterer.useFastClustering(cubeSize);
  before invoking the Cluster method.  
*/

template<class TransT, class ParticleT>
class RMSDCluster { 

private:
  //// Implementation Note: 
  //// We have defined this private class in order to avoid the need
  //// of the LEDA library in implementing the clustering phase.
  class Edge { 
  public:
    int vertex1, vertex2;
    float length;
    
    //// The default constructor
    Edge() : 
      vertex1(0), vertex2(0), length(0.0) { 
    }
    
    Edge(int vertex1_, int vertex2_, float length_):
      vertex1(vertex1_), vertex2(vertex2_), length(length_) {
    }
    
    friend bool operator<(const Edge& edge1, const Edge& edge2) {
      return (edge1.length < edge2.length);
    }
  };
  
  ////Private class intended to hold the TransT objects throughout the 
  // clustering process.
  class TransformationRecord {
  public:
    //used to indicate deletions in function work. Deletion will them 
    //be done faster with the clean method of RMSDCluster. Deletion 
    //in stl vectors is very slow.
    bool valid; 

    TransT trans;
    
    ////standard constructor.
    inline TransformationRecord(TransT& trans_): 
      valid(true), trans(trans_) {
    }
    
    virtual ~TransformationRecord() {}
    
    /// Join the tranfromations into this. 
    //  record is unchanged.
    virtual void join(TransformationRecord& record);

    inline float getScore() const {
      return trans.getScore();
    }
  };

public:
  ////default constructor
  // 
  RMSDCluster():isFast(false) {
  };
  
  // The main funcotion which should be called after the preProcessing 
  // is done.
  vector<TransT>& cluster(float maxDist, vector<TransT>& trans);

  //// preProcessing of a set of points in order to calculate distances 
  //on them.
  void preProcess(vector<ParticleT>& particles);


  //// Returns the RMSD of the preProcessed set of points under the two 
  // transformations.
  virtual float dist2(const TransT& trans1, const TransT& trans2);

  //// Sets the bin size on which the clustering will be done at the 
  // first stage. if not used the class will preform clustering on all 
  // bins at once.
  void useFastClustering(float binSize_);

  virtual ~RMSDCluster() {}

protected:

  //// The main loop builds the graph of transformations a contract 
  // the edges.
  virtual int work(float maxDist, vector<TransformationRecord *>& records);
  
  //// Just like work only handles each bin independently until there is 
  // nothing else to be done this way. Bins size is deremined by SetSpeed
  virtual int fastWork(float maxDist, vector<TransformationRecord *>& records);
  
  //// Remove transformations which are not valid.
  // should be used after each invokation of work.
  //virtual vector<TransformationRecord *> *clean(vector<TransformationRecord *>* records);
  virtual void clean(vector<TransformationRecord*>*& records);

private:
  bool isFast;
  float binSize;
  
  Molecule<ParticleT> molecule;

  // The controid of the molecule
  Vector3 centroid; 
};


/*********************************************
 * Template class RMSDCluster implementation *
 *********************************************/

template<class TransT, class ParticleT> void 
RMSDCluster<TransT, ParticleT>::TransformationRecord::join(RMSDCluster<TransT, ParticleT>::TransformationRecord& record) {
  trans.incScore(record.trans.getScore());
  trans.join(record.trans);
}

template<class TransT, class ParticleT> void 
RMSDCluster<TransT, ParticleT>::preProcess(vector<ParticleT>& particles){
  cout << "RMSDCluster: " << particles.size() 
       << " particles will be pre-processed" << endl;
  for (unsigned int i = 0 ; i < particles.size() ; i++) {
    molecule.add(particles[i]);
  }

  // Implementation Note: 
  // We keep the centroid, so we will not compute it more than once.
  centroid = molecule.centroid();
}

template <class TransT, class ParticleT> void 
RMSDCluster<TransT, ParticleT>::useFastClustering(float binSize_){ 
  binSize = binSize_; 
  isFast = true; 
  cout << "RMSDCluster: Bin size is set to " << binSize << endl;
}

template<class TransT, class ParticleT> float 
RMSDCluster<TransT, ParticleT>::dist2(const TransT& trans1, const TransT& trans2) {
  Molecule<ParticleT> molecule1 = molecule;
  molecule1 *= trans1.rigidTrans(); 

  Molecule<ParticleT> molecule2 = molecule;
  molecule2 *= trans2.rigidTrans();

  float rmsdDistance2 = 0;
  for (unsigned int i = 0 ; i < molecule.size() ; i++) {
    rmsdDistance2 += molecule1(i).dist2(molecule2(i));
  }

  return rmsdDistance2 / molecule.size();
}


/*
  One iteration of the clustering algorithm:
  create a graph of the transformations : each transformation is a vertex and the 
  weights on the edges are the distances between the transformations.
  Go through the edges in ascending lengths and every contract edges that are less than
  Maxdist long, but only use each vertex once. (throughout the iteration).

  For fast creation of the graph create a hash and put the picture of the centroid 
  by each transformation. Then calculate distance only between transformation which 
  the centroid is less than 2*Maxdist.
*/
template <class TransT, class ParticleT>
int RMSDCluster<TransT, ParticleT>::fastWork(float maxDist,
					     vector<TransformationRecord*>& records){
  float maxDist2 = maxDist * maxDist;
  int numOfJoins = 0;
  bool used[records.size()];
  
  GeomHash<Vector3, int> gHash(3, binSize);
  
  //load the hash
  for (unsigned int i = 0 ; i < records.size() ; ++i){
    used[i] = false;
    const RigidTrans3& rigidTrans = (records[i]->trans).rigidTrans();
    gHash.insert( rigidTrans * centroid, i);
  }
  
  cout << "RMSDCluster: Hash created" << endl;
  
  typedef vector<int> Bucket;
  vector<const Bucket *> *buckets = gHash.getBuckets();
  
  cout << "RMSDCluster: no. of buckets is : " << buckets->size() << endl;
  
  vector<const Bucket *>::iterator bucket;
  for (bucket = buckets->begin() ; bucket != buckets->end() ; ++bucket){
    
    // Implementation Note: edges vector is defined inside the loop so
    // that there will be a new vector for every bucket
    vector<Edge> edges;
    
    Bucket::const_iterator it1, it2;
    for (it1 = (*bucket)->begin() ; it1 != (*bucket)->end() ; ++it1) {
      for (it2 = it1 + 1 ; it2 != (*bucket)->end() ; ++it2) {
	float d2 = dist2(records[*it1]->trans, records[*it2]->trans);       
	if (d2 < maxDist2) {
	  edges.push_back(Edge(*it1, *it2, d2));
	}
      }
    }

    //cluster all transformations in the bin
    sort(edges.begin(), edges.end());
    
    vector<Edge>::iterator edge;
    for (edge = edges.begin(); edge != edges.end() ; edge++){
      //check if any end of the edge is deleted
      if (!used[edge->vertex1] && !used[edge->vertex2] && (edge->length < maxDist2)){
	numOfJoins++;
	used[edge->vertex1] = true;
	used[edge->vertex2] = true;

	
	TransformationRecord* record1 = records[edge->vertex1];  
	TransformationRecord* record2 = records[edge->vertex2];  
	if (record1->getScore() > record2->getScore()) {	
	  record1->join(*record2);
	  record2->valid = false;
	} else {
	  record2->join(*record1);
	  record1->valid = false;
	}
      }
    }
  }
  
  return numOfJoins;
}

template <class TransT, class ParticleT>
int RMSDCluster<TransT, ParticleT>::work(float maxDist,
					 vector<TransformationRecord *>& records) {
  float maxDist2 = maxDist * maxDist;
  int numOfJoins = 0;
  
  vector<Edge> edges;
  bool used[records.size()];
  
  GeomHash<Vector3, unsigned int> gHash(3, maxDist);
  
  for (unsigned int i =  0 ; i < records.size() ; ++i) {
    used[i] = false;

    const RigidTrans3& rigidTrans = (records[i]->trans).rigidTrans();
    gHash.insert(rigidTrans * centroid, i);
  }
   
  cout << "RMSDCluster: Hash created" << endl;
  
  HashResult<unsigned int>::iterator it;
  for (unsigned int i = 0 ; i < records.size() ; ++i) {
    HashResult<unsigned int> result;

    const RigidTrans3& rigidTrans = (records[i]->trans).rigidTrans();
    gHash.query(rigidTrans * centroid, maxDist, result);
    
    for (it = result.begin() ; it != result.end() ; ++it) {
      // only create an edge if i < *it so that 
      // an edge will not be created twice
      if (i < *it) {
	float d2 = dist2(records[i]->trans, records[*it]->trans);
	if (d2 < maxDist2) {
	  edges.push_back(Edge(i, *it, d2));
	}
      }
    }
  }
  
  sort(edges.begin(), edges.end());
  
  vector<Edge>::iterator edge;
  for (edge = edges.begin() ; edge != edges.end() ; edge++){
    //check if any end of the edge is deleted
    // To Check if this condition is necessary
    if (!used[edge->vertex1] && !used[edge->vertex2] && edge->length < maxDist2){
      numOfJoins++;

      used[edge->vertex1] = true;
      used[edge->vertex2] = true;

      records[edge->vertex1]->join(*(records[edge->vertex2]));
      records[edge->vertex2]->valid = false;
    }
  }
  
  return numOfJoins;
}

// template <class TransT, class ParticleT> vector<RMSDCluster<TransT, ParticleT>::TransformationRecord *>* RMSDCluster<TransT, ParticleT>::clean(vector<TransformationRecord *>* records){
//   vector<TransformationRecord *> *results = new vector<TransformationRecord *>();

//   for (unsigned int i = 0 ; i < records->size() ; i++){
//     if ((*records)[i]->valid) {
//       results->push_back((*records)[i]);
//     } else {
//       delete((*records)[i]);
//     }
//   }
  
//   return results;
// }


template <class TransT, class ParticleT> void RMSDCluster<TransT, ParticleT>::clean(vector<TransformationRecord*>*& records) {
  vector<TransformationRecord*> *results = new vector<TransformationRecord*>();

  for (unsigned int i = 0 ; i < records->size() ; i++){
    if ((*records)[i]->valid) {
      results->push_back((*records)[i]);
    } else {
      delete((*records)[i]);
    }
  }
  
  records->clear();
  records = results;
}





template <class TransT, class ParticleT> vector<TransT>& 
RMSDCluster<TransT, ParticleT>::cluster(float maxDist, 
					vector<TransT>& trans){
  cout << "RMSDCluster: Num of trans = " << trans.size() << endl;

  //create initial vectors of transformation records and bit vector to
  //indicate what is deleted
  vector<TransformationRecord*>* records = new vector<TransformationRecord*>();
  
  for ( vector<TransT>::iterator iter = trans.begin(), 
	  endIter = trans.end() ; iter != endIter ; ++iter){
    TransformationRecord* record = new TransformationRecord(*iter);
    records->push_back(record);
  }
  
  //the main loop
  if (isFast){
    cout << "." << endl;
    //use fastWork first
    while (fastWork(maxDist, *records)){
      clean(records);
    }
  }
  
  while (work(maxDist, *records)){
    cout << "." << endl;
    clean(records);
  }
  
  
  //build the vector for output
  vector<TransT> *results = new vector<TransT>();
  for (unsigned int i = 0 ; i < records->size() ; ++i){
    results->push_back((*records)[i]->trans);
    delete((*records)[i]);
  }

  delete(records);

  cout << "RMSDCluster: leaving cluster" << endl;
  return *results;
}

#endif //__RMSDCLUSTER_H

















