#ifndef __SecondaryStructure_H
#define __SecondaryStructure_H


#include "Helix.h"
#include "Sheet.h"
#include "SmallVector.h"
#include "Atom.h"
#include <vector.h>
#include <iostream.h>
#include <pair.h>

/*
CLASS 
  SecondStruct  
  
  This class encapsulate all the reading of the secondary structure 
  from the PDB file and gives various methods of verification on them

KEWORDS
  secondary, structure, alpha, helix, beta, strand, sheet, connection,
  molecule, atom, PDB

AUTHORS
  Zipi Fligelamn (zipo@math.tau.ac.il)

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1999.

GOALS
  The goals of this class is save for the user the need to anderstand all
  the particulars of the representaion of the secondary structures in the 
  PDB and yet allow him to use its infromation. The class has various quering
  methods

USAGE
*/
class SecondStruct 
{
public:
  typedef SmallVector<Helix> Helices;
  typedef SmallVector<Sheet> Sheets;
  typedef enum { Unknown=0, AHelix, BSheet, Turn} SSType; 
  
  // GROUP: Constructors
  
  //// Empty Constructor
  SecondStruct();
  
  //// A constructor that will read all the relevand infromation from
  // the PDB file and then close it
  explicit SecondStruct(const char* const PDBFileName); 

    //// Copy constructor
    SecondStruct(const SecondStruct& ss);

  // GROUP: Data members infromation
  
  //// The small array of Helices
  inline const Helices& AlphaHelices() const;
  
  //// The small array of BetaSheets
  inline const Sheets&  BetaSheets() const;
  
  // GROUP: Inspecting particular information
  
  //// Requesting a certain helix by its id.
  Helix& operator[](const int helixId);
  
  //// Requesting a certain sheet by its letter id. If it not found
  // the end od the small vector is returned
  // The user need to check this out before proceeding
  // A particular strand can afterwood be requested since the BetaSheet
  // class inherits from SmallArray where a operator[] is defined
  Sheet& operator[](const char strandId);
  
  // GROUP: Atoms belogning inspections

  //// Checking whether a certain atom in a certain chain is 
  // located in a particular alpha helix
  bool isInAHelix(const short helixNum, const int atomNum) const;
  
  //// Checking whether a certain atom in a certain chain is 
  // located in a particular beta sheet
  // Since this is done by checking all the strands untild finding the
  // strand that the atoms reside in this act might be computationally 
  // expensive
  bool isInBSheet(const char sheetId,  const int atomNum) const;
  
  //// given a vector of indecises will return true if the indices
  // come from the same SS structure or not 
  // at the moment atoms that come from the same sheet but not
  // from the same strand are considered as from different SS
  bool isFromSameSS(const vector<short> &indices);
  
  //// same as above only with a regular array.
  bool isFromSameSS(const short indices[], int len);

  //// Making a general search on all the secondary structures
  // returning the type if the atom was found in one of them and 
  // known if not
  SSType atomInSS(const int atomNum) const;
  
  //// Looking for the atom in all the helices and returning the 
  // helix number if found, and -1 if not.
  short numOfAHelix(const int atomNum) const;

  //// Looking for the atom in all the sheets and returning
  // the char id of the sheet if found, and '\0' if not
  char numOfBSheet(const int atomNum) const;

   //// Looking for the atom in all the sheets and returning a pair
  // the char id of the sheet if found, and '\0' if not, as the first
  // argument. The strand number in case there is a sheet and -1 if not.
  pair<char,short> sheetAndStrandId(const int atomNum) const;

  
  //// Listing of all the secondary structure in the printing format
  // defined inside Helix, Strand and BetaSheet classes
  friend ostream& operator<<(ostream& s, const SecondStruct& sst);
private:
  void readSSFromPDBFile(ifstream& PDBstream);
 
  // data members
  Helices alpha;
  Sheets beta;
};

#endif 

/*************************
 Inline definitions
 ************************/
const SecondStruct::Helices& SecondStruct::AlphaHelices() const
{
  return alpha;
}

const SecondStruct::Sheets& SecondStruct::BetaSheets() const
{
  return beta;
}
