#ifndef SeqScore_hpp
#define SeqScore_hpp

#include "AminoAcid.h"
#include "Atom.h"
#include "Match.h"

template<class PDBrec= vector<Atom>,class Mtch= Match >

/*
CLASS
  SeqScore
   
  This class serves for a score calculation of sequence of amino acid pairs.
  The score can be based on the PAM250 matrix or any other user defined score matrix.

KEYWORD
    Atom, Match, ScoreMatrix

AUTHORS
    Maxim Shatsky (maxshats@math.tau.ac.il)

    copyright: SAMBA Group , Tel-Aviv Univ. Israel, 1997, 1999.

CHANGES LOG


USAGE
 Before calculating the score of two matched amino acid sequences,the user
 should define a score matrix.This is a 20x20 matrix M when M[i,j] which 
 defines a value for AminoAcid_i against AminoAcid_j (0<=i,j<=19). 
 The amino acids are indexed in the following way (see AminoAcid class for enum RESIDUE_INDEX): 
  ARG,LYS,GLU,GLN,ASP,ASN,PRO,GLY,SER,
  THR,HIS,ALA,ILE,LEU,VAL,PHE,CYS,MET,TYR,TRP. (ARG=0,...,TRP=19)
 It means that M[3,0]=M[GLN,ARG] (to map amino acid to proper index see AminoAcid class) .
 There is a predefined PAM250 score matrix SeqScore<>::PAM250 ,someone
 can use it as a template for creating another score matrix. 

 EXAMPLE
 
 vector<Atom> vModel;
 vector<Atom> vTarget;
 //...
 //fill atom information into vectors vModel,vTarget  
 //...
 vector<Match*> vMatches; 
 //..
 //find some matches 
 //...

 SeqScore<>::PAM250 pam250; //define score matrix
 SeqScore<> seqScore;

 //calculate and output PAM250 score for each found match. 
 for(int i=0;i < vMatches.size();i++){
  float score=seqScore.calcScore(vModel,
                          vTarget,
                          *vMatches[i],
                          pam250);
  score=score/vMatches[i].size(); //normolize the score
  cout<<score<<endl;
 }

 END

 SeqScore is a template that is defined by 2 classes. 
 EXAMPLE
 SeqScore<class PDBrec= vector<Atom>,class Mtch= Match> seqScore;
 END

 Class PDBrec should support the following:
 EXAMPLE
 PDBrec pdbRec;
 pdbRec[int index].residueType(); //this function returns one-letter abbreviation of amino acid residue.
 END
 
 Class Mtch should support the following:
 EXAMPLE
 Mtch match;
 match[i].model; //returns index in  pdbRec
 match[i].scene; //returns index in  pdbRec
 END

 If one wishes to create its own score matrix just see how PAM250 class
 is defined.
*/



class SeqScore
{

public:
 


  ////ScoreMatrix class defines the score for each pair of amino acid residues
  class ScoreMatrix{
  protected:
    float m_fScore[20][20];
  public:
    ////Returns the score
    float score(int i,int j)const;
  };


   ////Returns the score of two matched sequences.Complexity O(match.size()).
  float calcScore( const PDBrec& pdbModel,
                   const PDBrec& pdbTarget,
                   const Mtch& match,
                   const ScoreMatrix& m_scoreMtrx)const{
    float fSum=0;

    for(int i=0;i<match.size();i++){
      RESIDUE_INDEX model=AminoAcid::getResidueTypeIndx(pdbModel[match[i].model].residueType());
      RESIDUE_INDEX target=AminoAcid::getResidueTypeIndx(pdbTarget[match[i].scene].residueType());
      
      fSum+= m_scoreMtrx.score( model , target );
    }
    
    return fSum;
  }

  ////Extends ScoreMatrix. The score is defined in the following way:
  //Hydrophobic matched with Hydrophobic receives 1.
  //Hydrophilic matched with Hydrophilic receives 1.
  //Otherwise -1.
  class Hydrophobic:public ScoreMatrix{
  public:
    ////Constructor. Complexity O(400).
    Hydrophobic();
  };

  ////Extends ScoreMatrix. The score is defined as in PAM250 matrix.
  class PAM250:public ScoreMatrix{
  public:
    ////Constructor. Complexity O(400).
    PAM250();
  };
};

template<class PDBrec,class Mtch>
float SeqScore<PDBrec,Mtch>::ScoreMatrix::score(int i,int j)const{return m_fScore[i][j];};

template<class PDBrec,class Mtch>
SeqScore<PDBrec,Mtch>::PAM250::PAM250(){
    m_fScore[ARG][ARG]=6;
    m_fScore[ARG][LYS]=3;
    m_fScore[ARG][GLU]=-1;
    m_fScore[ARG][GLN]=1;
    m_fScore[ARG][ASP]=-1;
    m_fScore[ARG][ASN]=0;
    m_fScore[ARG][PRO]=0;
    m_fScore[ARG][GLY]=-3;
    m_fScore[ARG][SER]=0;
    m_fScore[ARG][THR]=-1;
    m_fScore[ARG][HIS]=2;
    m_fScore[ARG][ALA]=-2;
    m_fScore[ARG][ILE]=-2;
    m_fScore[ARG][LEU]=-3;
    m_fScore[ARG][VAL]=-2;
    m_fScore[ARG][PHE]=-4;
    m_fScore[ARG][CYS]=-4;
    m_fScore[ARG][MET]=0;
    m_fScore[ARG][TYR]=-4;
    m_fScore[ARG][TRP]=2;

    m_fScore[LYS][LYS]=5;
    m_fScore[LYS][GLU]=0;
    m_fScore[LYS][GLN]=1;
    m_fScore[LYS][ASP]=0;
    m_fScore[LYS][ASN]=1;
    m_fScore[LYS][PRO]=-1;
    m_fScore[LYS][GLY]=-2;
    m_fScore[LYS][SER]=0;
    m_fScore[LYS][THR]=0;
    m_fScore[LYS][HIS]=0;
    m_fScore[LYS][ALA]=-1;
    m_fScore[LYS][ILE]=-2;
    m_fScore[LYS][LEU]=-3;
    m_fScore[LYS][VAL]=-2;
    m_fScore[LYS][PHE]=-5;
    m_fScore[LYS][CYS]=-5;
    m_fScore[LYS][MET]=0;
    m_fScore[LYS][TYR]=-4;
    m_fScore[LYS][TRP]=-3;
    
    m_fScore[GLU][GLU]=4;
    m_fScore[GLU][GLN]=2;
    m_fScore[GLU][ASP]=3;
    m_fScore[GLU][ASN]=1;
    m_fScore[GLU][PRO]=-1;
    m_fScore[GLU][GLY]=0;
    m_fScore[GLU][SER]=0;
    m_fScore[GLU][THR]=0;
    m_fScore[GLU][HIS]=1;
    m_fScore[GLU][ALA]=0;
    m_fScore[GLU][ILE]=-2;
    m_fScore[GLU][LEU]=-3;
    m_fScore[GLU][VAL]=-2;
    m_fScore[GLU][PHE]=-5;
    m_fScore[GLU][CYS]=-5;
    m_fScore[GLU][MET]=-2;
    m_fScore[GLU][TYR]=-4;
    m_fScore[GLU][TRP]=-7;

    m_fScore[GLN][GLN]=4;
    m_fScore[GLN][ASP]=2;
    m_fScore[GLN][ASN]=1;
    m_fScore[GLN][PRO]=0;
    m_fScore[GLN][GLY]=-1;
    m_fScore[GLN][SER]=-1;
    m_fScore[GLN][THR]=-1;
    m_fScore[GLN][HIS]=3;
    m_fScore[GLN][ALA]=0;
    m_fScore[GLN][ILE]=-2;
    m_fScore[GLN][LEU]=-2;
    m_fScore[GLN][VAL]=-2;
    m_fScore[GLN][PHE]=-5;
    m_fScore[GLN][CYS]=-5;
    m_fScore[GLN][MET]=-1;
    m_fScore[GLN][TYR]=-4;
    m_fScore[GLN][TRP]=-5;
 
    m_fScore[ASP][ASP]=4;
    m_fScore[ASP][ASN]=2;
    m_fScore[ASP][PRO]=-1;
    m_fScore[ASP][GLY]=1;
    m_fScore[ASP][SER]=0;
    m_fScore[ASP][THR]=0;
    m_fScore[ASP][HIS]=1;
    m_fScore[ASP][ALA]=0;
    m_fScore[ASP][ILE]=-2;
    m_fScore[ASP][LEU]=-4;
    m_fScore[ASP][VAL]=-2;
    m_fScore[ASP][PHE]=-6;
    m_fScore[ASP][CYS]=-5;
    m_fScore[ASP][MET]=-3;
    m_fScore[ASP][TYR]=-4;
    m_fScore[ASP][TRP]=-7;
    
    m_fScore[ASN][ASN]=2;
    m_fScore[ASN][PRO]=-1;
    m_fScore[ASN][GLY]=0;
    m_fScore[ASN][SER]=1;
    m_fScore[ASN][THR]=0;
    m_fScore[ASN][HIS]=2;
    m_fScore[ASN][ALA]=0;
    m_fScore[ASN][ILE]=-2;
    m_fScore[ASN][LEU]=-3;
    m_fScore[ASN][VAL]=-2;
    m_fScore[ASN][PHE]=-3;
    m_fScore[ASN][CYS]=-4;
    m_fScore[ASN][MET]=-2;
    m_fScore[ASN][TYR]=-2;
    m_fScore[ASN][TRP]=-4;

    m_fScore[PRO][PRO]=6;
    m_fScore[PRO][GLY]=-1;
    m_fScore[PRO][SER]=1;
    m_fScore[PRO][THR]=0;
    m_fScore[PRO][HIS]=0;
    m_fScore[PRO][ALA]=1;
    m_fScore[PRO][ILE]=-2;
    m_fScore[PRO][LEU]=-3;
    m_fScore[PRO][VAL]=-1;
    m_fScore[PRO][PHE]=-5;
    m_fScore[PRO][CYS]=-3;
    m_fScore[PRO][MET]=-2;
    m_fScore[PRO][TYR]=-5;
    m_fScore[PRO][TRP]=-6;

    m_fScore[GLY][GLY]=5;
    m_fScore[GLY][SER]=1;
    m_fScore[GLY][THR]=0;
    m_fScore[GLY][HIS]=-2;
    m_fScore[GLY][ALA]=1;
    m_fScore[GLY][ILE]=-3;
    m_fScore[GLY][LEU]=-4;
    m_fScore[GLY][VAL]=-1;
    m_fScore[GLY][PHE]=-5;
    m_fScore[GLY][CYS]=-3;
    m_fScore[GLY][MET]=-3;
    m_fScore[GLY][TYR]=-5;
    m_fScore[GLY][TRP]=-7;

    m_fScore[SER][SER]=2;
    m_fScore[SER][THR]=1;
    m_fScore[SER][HIS]=-1;
    m_fScore[SER][ALA]=1;
    m_fScore[SER][ILE]=-1;
    m_fScore[SER][LEU]=-3;
    m_fScore[SER][VAL]=-1;
    m_fScore[SER][PHE]=-3;
    m_fScore[SER][CYS]=0;
    m_fScore[SER][MET]=-2;
    m_fScore[SER][TYR]=-3;
    m_fScore[SER][TRP]=-2;
    
    m_fScore[THR][THR]=3;
    m_fScore[THR][HIS]=-1;
    m_fScore[THR][ALA]=1;
    m_fScore[THR][ILE]=0;
    m_fScore[THR][LEU]=-2;
    m_fScore[THR][VAL]=0;
    m_fScore[THR][PHE]=-3;
    m_fScore[THR][CYS]=-2;
    m_fScore[THR][MET]=-1;
    m_fScore[THR][TYR]=-3;
    m_fScore[THR][TRP]=-5;
    
    m_fScore[HIS][HIS]=6;
    m_fScore[HIS][ALA]=-1;
    m_fScore[HIS][ILE]=-2;
    m_fScore[HIS][LEU]=-2;
    m_fScore[HIS][VAL]=-2;
    m_fScore[HIS][PHE]=-2;
    m_fScore[HIS][CYS]=-3;
    m_fScore[HIS][MET]=-2;
    m_fScore[HIS][TYR]=0;
    m_fScore[HIS][TRP]=-3;

    m_fScore[ALA][ALA]=2;
    m_fScore[ALA][ILE]=-1;
    m_fScore[ALA][LEU]=-2;
    m_fScore[ALA][VAL]=0;
    m_fScore[ALA][PHE]=-3;
    m_fScore[ALA][CYS]=-2;
    m_fScore[ALA][MET]=-1;
    m_fScore[ALA][TYR]=-3;
    m_fScore[ALA][TRP]=-6;

    m_fScore[ILE][ILE]=5;
    m_fScore[ILE][LEU]=2;
    m_fScore[ILE][VAL]=4;
    m_fScore[ILE][PHE]=1;
    m_fScore[ILE][CYS]=-2;
    m_fScore[ILE][MET]=2;
    m_fScore[ILE][TYR]=-1;
    m_fScore[ILE][TRP]=-5;

    m_fScore[LEU][LEU]=6;
    m_fScore[LEU][VAL]=2;
    m_fScore[LEU][PHE]=2;
    m_fScore[LEU][CYS]=-6;
    m_fScore[LEU][MET]=4;
    m_fScore[LEU][TYR]=-1;
    m_fScore[LEU][TRP]=-2;

    m_fScore[VAL][VAL]=4;
    m_fScore[VAL][PHE]=-1;
    m_fScore[VAL][CYS]=-2;
    m_fScore[VAL][MET]=2;
    m_fScore[VAL][TYR]=-2;
    m_fScore[VAL][TRP]=-6;

    m_fScore[PHE][PHE]=9;
    m_fScore[PHE][CYS]=-4;
    m_fScore[PHE][MET]=0;
    m_fScore[PHE][TYR]=7;
    m_fScore[PHE][TRP]=0;

    m_fScore[CYS][CYS]=12;
    m_fScore[CYS][MET]=-5;
    m_fScore[CYS][TYR]=0;
    m_fScore[CYS][TRP]=-8;

    m_fScore[MET][MET]=6;
    m_fScore[MET][TYR]=-2;
    m_fScore[MET][TRP]=-4;

    m_fScore[TYR][TYR]=10;
    m_fScore[TYR][TRP]=0;

    m_fScore[TRP][TRP]=17;

    //complete semetry matrix 
    for(int i=0;i<20;i++)
      for(int j=i;j<20;j++)
        m_fScore[j][i]=m_fScore[i][j];
      
}

template<class PDBrec,class Mtch>
SeqScore<PDBrec,Mtch>::Hydrophobic::Hydrophobic(){
  
      //Surface
      for(int i=0;i<11;i++)
        for(int j=i;j<11;j++)
          m_fScore[i][j]=1;
      
      //Hydrophobic
      for(int i=11;i<20;i++)
        for(int j=i;j<20;j++)
          m_fScore[i][j]=1;
      
      //Hydrophobic with Surface
      for(int i=0;i<11;i++)
        for(int j=11;j<20;j++)
          m_fScore[i][j]=-1;
      
      //complete semetry matrix 
      for(int i=0;i<20;i++)
        for(int j=i;j<20;j++)
          m_fScore[j][i]=m_fScore[i][j];
}
#endif





