#ifndef _SmallVector_h
#define _SmallVector_h


class ShortVectorBase
{
public:
  typedef unsigned short size_type;

  size_type size() const;
  size_type space() const;
  void incrAllocSpace();

protected:
  size_type alloc;
  size_type sz;
};


/*
CLASS
  SmallVector

  Defines a vector template. This class template was designed as a cheap
  substitute to vector<> class template. The SmallVector class has a smaller
  executable size, takes up less memory, is simpler and should usually prove
  to be faster than the larger, more verbose vector<> class.

AUTHORS
  Meir Fuchs. (meirfux@math.tau.ac.il)

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1997.

CHANGES LOG
<UL>
<LI>
</UL>

GOALS
  Defines a cheap vector<> template class. The size of the vector may be
  limited, no insertion methods only push_back, simple deletion methods. There
  is no allocator class involved and the iterators are defined as simple
  pointers. This results in smaller excutable footprint, some reduction in
  memory usage and usually faster code.

USAGE
  Usage of the SmallVector class template is similar to that of the vector
  class template.
 
  Note that one may supply a different base class to the SmallVector class.
  This immediately raises two possiblilities:<BR>
  a. Change size and space to integers and thereby allow virtually unlimited
  vector length.<BR>
  b. Redefine the incrAllocSpace function which calculates the amount of space
  to be allocated when the allocated space overflows due to a push_back 
  operation.

  The class is not truly STL compliant though feeding this container into some
  of the STL algoritms might actually work.
*/
template<class T, class Alloc = ShortVectorBase>
class SmallVector : 
  public ShortVectorBase
{
public:

  //// STL compliant typedef.
  typedef T value_type;
  //// STL compliant typedef.
  typedef typename Alloc::size_type size_type;
  //// STL compliant typedef.
  typedef T* iterator;
  //// STL compliant typedef.
  typedef const T* const_iterator;
  //// STL compliant typedef.
  typedef T& reference;
  //// STL compliant typedef.
  typedef const T& const_reference;

  //// Constructor receives initial space to allocate. Default is 3.
  explicit SmallVector(unsigned short defAlloc=3);

    //// Constructor receives initial number of objects in create in the
    //// vector and initial space to allocate. 
    //// If space is too small to fit objects - SmallVector will allocate more!
    //// Please VERY be carefull the first object are NOT initiated they are 
    //// meaningless objects 
  explicit SmallVector(unsigned short Objects, unsigned short defAlloc=3);

  //// Copy constructor
  SmallVector(const SmallVector<T, Alloc>& a);

  //// Destructor. Deletes vector data.
  ~SmallVector();

  SmallVector<T, Alloc>& operator=(const SmallVector<T, Alloc>& a);

  //// Adds an item at the end of the array. Allocates more space if needed.
  void push_back(const T& elem);

  //// Returns reference to indexed element of the vector.
  reference operator[](unsigned short index);

  //// Returns a const reference to indexed element of the vector.
  const_reference operator[](unsigned short index) const;

  //// Returns a pointer to the vector's head. 
  iterator begin();

  //// Returns a pointer to the vector's head. 
  const_iterator begin() const;

  //// Returns a pointer to the vector's end. The end is past the last element
  // of the vector.
  iterator end();

  //// Returns a pointer to the vector's end. The end is past the last element
  // of the vector.
  const_iterator end() const;
  
private:
  T* data;

  static T* allocate(unsigned short space);
  static void copy(T* start, T* end, T* target);
  static void destroy(T* start, T* end);
};

/******************
 * implementation *
 ******************/  

inline unsigned short ShortVectorBase::size() const
{
  return sz;
}

inline unsigned short ShortVectorBase::space() const
{
  return alloc;
}


template<class T, class Alloc>
T* SmallVector<T, Alloc>::allocate(unsigned short space)
{
  return (T*)::operator new(space*sizeof(T));
}

template<class T, class Alloc>
void SmallVector<T, Alloc>::copy(T* start, T* end, T* target)
{
  for (; start != end; ++start, ++target)
    *target = *start;
}

template<class T, class Alloc>
void SmallVector<T, Alloc>::destroy(T* start, T* end)
{
  for (T* ptr=start; ptr!=end; ++ptr)
    ptr->~T();
  ::operator delete(start);
}


template<class T, class Alloc>
SmallVector<T, Alloc>::SmallVector(unsigned short Objects, unsigned short defAlloc)
{
    sz=Objects;
    alloc = (defAlloc > Objects)? defAlloc : Objects;
    data = allocate(alloc);
}

template<class T, class Alloc>
SmallVector<T, Alloc>::SmallVector(unsigned short defAlloc/*=3*/)
{
  sz = 0;
  alloc = (defAlloc == 0) ? 1 : defAlloc; 
  data = allocate(alloc);
}



template<class T, class Alloc>
SmallVector<T, Alloc>::SmallVector(const SmallVector<T, Alloc>& a)
{
  alloc = sz = a.sz;
  data = allocate(alloc);
  copy(a.data, a.data+sz, data);
}

template<class T, class Alloc>
SmallVector<T, Alloc>::~SmallVector()
{
  destroy(data, data+sz);
}
  
template<class T, class Alloc>
SmallVector<T, Alloc>& SmallVector<T, Alloc>::operator=(const SmallVector<T, Alloc>& a)
{
  if (&a != this) {
    destroy(data, data+sz);
    alloc = sz = a.sz;
    data = allocate(alloc);
    copy(a.data, a.data+sz, data);
  }
  return *this;
}  

template<class T, class Alloc>
void SmallVector<T, Alloc>::push_back(const T& elem)
{
  if (sz>=alloc) {
    T* temp = data;
    incrAllocSpace();
    data = allocate(alloc);
    copy(temp, temp+sz, data);
    destroy(temp, temp+sz);
  }
  data[sz] = elem;
  ++sz;
}
   
//// An operator that given an index i returns the SmallVector[i]
template<class T, class Alloc>
T& SmallVector<T, Alloc>::operator[](unsigned short index)
{
  return data[index];
}

//// The same however deals with a const vector therefore returns a const 
// element
template<class T, class Alloc>
const T& SmallVector<T, Alloc>::operator[](unsigned short index) const
{
  return data[index];
}


template<class T, class Alloc>
T* SmallVector<T, Alloc>::begin()
{
  return data;
}

template<class T, class Alloc>
T* SmallVector<T, Alloc>::end()
{
  return data+sz;
}

template<class T, class Alloc>
const T* SmallVector<T, Alloc>::begin() const
{
  return data;
}

template<class T, class Alloc>
const T* SmallVector<T, Alloc>::end() const
{ 
  return data+sz; 
}

#endif









