#include "Strand.h"

typedef StrandConnection::ConnectEnd ConnectEnd;

StrandConnection::StrandConnection()
  : curr(ConnectEnd('?',-1)), prev(ConnectEnd('?',-1)) 
{}

StrandConnection::StrandConnection(const char curAtomId, const int curResSeq,
                 const char prevAtomId, const int prevResSeq)
  : curr(ConnectEnd(curAtomId,curResSeq)),
      prev(ConnectEnd(prevAtomId,prevResSeq))  
{}

StrandConnection::StrandConnection(const StrandConnection& stcn)
  : curr(stcn.curr), prev(stcn.prev)
{}

StrandConnection::StrandConnection(const char* const PDBrec)
  : curr(ConnectEnd(StrandPDB::currAtom(PDBrec), 
                    StrandPDB::currResSeq(PDBrec))),
    prev(ConnectEnd(StrandPDB::prevAtom(PDBrec), 
                    StrandPDB::prevResSeq(PDBrec))) 
{} 

Vector3 StrandConnection::connectionDirection(const Molecule<Particle>& mol) const
{
  Vector3 dir = mol[curr.second].position() - mol[prev.second].position();
  dir/= dir.norm();
  return dir;
}

bool StrandConnection::operator==(const StrandConnection& stcn)
{
  return ( (curr.first == stcn.curr.first) &&
           (curr.second == stcn.curr.second) &&
           (prev.first == stcn.prev.first) &&
           (prev.second == stcn.prev.second));
}

bool StrandConnection::isEmpty() const
{
  return ((curr.first == '?') && (curr.second == -1) &&
          (prev.first == '?') && (prev.second == -1));
}

ostream& operator<<(ostream& s, const StrandConnection& stcn)
{
  s << "from atom (" << stcn.prev.first << "," << stcn.prev.second << ") " 
    << "  to atom (" << stcn.curr.first << "," << stcn.curr.second << ") "
    << endl;
  return s;
}

typedef Strand::RelativeParallelism RelativeParallelism;

Strand::Strand()
  :strandNum(-1), initResNum(-1),endResNum(-1), initChainId('?'),
   endChainId('?'), pr(0), cnt(StrandConnection())
{}

Strand::Strand(short StrandNum, int initResidueNum, int endResidueNum, 
               char initChainID, char endChainID, 
               RelativeParallelism sense, const StrandConnection& stcn)
  : strandNum(StrandNum), initResNum(initResidueNum),
    endResNum(endResidueNum), initChainId(initChainID),
    endChainId(endChainID), pr(sense), cnt(stcn)
{}

Strand::Strand(const char* const PDBrec)
  : strandNum(StrandPDB::strandNum(PDBrec)), 
    initResNum(StrandPDB::initSeqNum(PDBrec)),
    endResNum(StrandPDB::endSeqNum(PDBrec)),
    initChainId(StrandPDB::initChainId(PDBrec)),
    endChainId(StrandPDB::endChainId(PDBrec)),
    pr(StrandPDB::sense(PDBrec)),
    cnt(PDBrec)
{}

Strand::Strand(const Strand& st)
    : strandNum(st.strandNum), initResNum(st.initResNum),
      endResNum(st.endResNum), initChainId(st.initChainId),
      endChainId(st.endChainId), pr(st.pr), cnt(st.cnt)
{}

bool Strand::isAtomInStrand(const Atom& at) const 
{ 
    char chain = at.chainId();
    //// if atom has no chain info and strand has no chain info 
    // compare residue index
    if ((chain == ' ' && initChainId != '?') ||
        //// make sure the chains are correct 
        (chain == initChainId && chain == endChainId))
        return isResInStrand(at.residueIndex());
    else
        return false;
}

ostream& operator<<(ostream& s, const Strand& str)
{
  s << "STRAND num:" << str.strandNum << endl;
  s << "begin: (" << str.initResNum << " , " << str.initChainId << ") "
    << "end: (" << str.endResNum << " , " << str.endChainId <<")\n";
  s << "Type: ";
  if (str.pr == 0)
    s << "First Strand";
  else if (str.pr == 1)
    s << "Parallel";
  else if (str.pr == -1)
    s << "Anti Parallel";
  else {
    cerr << "fatal error occured unknow type of beta strand\n";
    exit(1);
  }
  s << endl;
  s << str.cnt;
  return s;
}







