/* $Id: Vector3.h,v 1.1 1997/08/10 11:54:04 meirfux Exp $ */

#ifndef _Vector3_h

#define _Vector3_h

#include "iostream.h"

/* 
CLASS 
  Vector3

  Defines a 3D vector class with operators.

KEYWORDS
  linear, algebra, vector, angle, distance, norm, real

AUTHORS
  Meir Fuchs. (meirfux@math.tau.ac.il)

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 1997.

CHANGES LOG
<UL>
<LI>
</UL>

GOALS
  Since most of the work on protein structure is done in 3D space, a fast,
  well designed 3D vector class was in place. Knowing the dimension is 3
  allows for a fast, loopless implemetation of Vector3 and Matrix3 to be 
  written.
  
USAGE
  Many 3D vector operations are supplied including vector addition and 
  subtraction, dot-product mult., multiplication by scalar, distance and 
  angular operators.<P>

  A new Vector3 may be constructed using 3 floats or an array of 3 floats.
  EXAMPLE
    Vector3 v(1.0, 2.0, 3.0);
    float x[3];
    Vector3 v(x);
  END

  Coordinates may be accessed using the x(int) method which returns the
  vector's given coordinate (1..3) or using the [] operator (0..2).
  EXAMPLE
    if (v[0] == v.x(1))  cout << "OK\n";
  END

  Addition, subtraction, multiplication and division by scalar, dot-product
  distance, angle.
  EXAMPLE
    w = u+v;                           // addition
    w-= v;                             // in-place subtraction w==u
    if ((w|u) == 0) cout << "OK\n";    // distance operator |
    if (w^u == 0) cout << "OK\n";      // angle operator ^
    w=2*w;                             // multiplicatio by scalar
    float f=w*v;                       // dot-product;
  END

*/
class Vector3
{

public:
  // GROUP: define real type.

  //// Defines the type with which coordinates are defined. Changing this to
  // double will casue the Vector3 and Matrix3 classes to work with double
  // precision coordinates. float was chosen for considerations of speed.
  // double precision is rarely needed since strctural data is inexact.
  typedef float real;

  // GROUP: Constructors.

  //// Initialize to the 0 vector.
  Vector3();

  //// Initialize using 3 coordinates x,y,z.
  Vector3(const real& nx, const real& ny, const real& nz);

  //// initialize using 3 coordinates in array. 
  Vector3(const real x[3]);

  //// initialize using 3 double precision coordinates in array. 
  Vector3(const double x[3]);
  
  // GROUP: Inspection.

  //// return the x, y or z coordinates. 
  real x() const;
  real y() const;
  real z() const;

  //// Returns x=x(1), y=x(2) or z=x(3) coordinate. If other values are given
  // result are unexpected. 
  real x(const unsigned short coord) const;

  
  // Group Update
  void updateX(const real& x);
  void updateY(const real& y);
  void updateZ(const real& z);

  //// Return Vector3's distance from origin.
  real norm() const;

  //// Return Vector3's squared distance ffrom origin.
  real norm2() const;

  //// Returns Vector3's distance from Vector3 p.
  real dist(const Vector3 &p) const;

  //// Returns Vector3's distance squared from Vector3 p.
  real dist2(const Vector3 &p) const;

  //// Treats vector like a real array. Acceptable indices 0, 1, 2. Other
  // indices may cause run-time errors. 
  const real& operator[](const unsigned short coord) const{
    return c[coord];
  }

  // GROUP: Operators.

  //// Returns Vector3 addition of two Vector3s.
  friend Vector3 operator+(const Vector3 &, const Vector3 &);

  //// Returns Vector3 subtraction of two Vector3s.
  friend Vector3 operator-(const Vector3 &, const Vector3 &);

  //// Returns Vector3 negative.
  friend Vector3 operator-(const Vector3 &);

  //// Returns dot product of two Vector3s.
  friend real operator*(const Vector3 &, const Vector3 &);

  //// Returns vector multiplied by scalar m.
  friend Vector3 operator*(const Vector3 &, const real &);

  //// Returns vector multiplied by scalar m.
  friend Vector3 operator*(const real &, const Vector3 &);

  //// Returns vector divided by scalar m.
  friend Vector3 operator/(const Vector3 &, const real &);

  //// Returns vector that is vertical to both vectors.
  friend Vector3 operator&(const Vector3 &, const Vector3 &);

  //// Returns distance between 2 Vector3s (similar to dist).
  friend real operator|(const Vector3 &, const Vector3 &);

  //// Returns angle between p1 and p2 in radians.
  friend real operator^(const Vector3 &, const Vector3 &);

  //// Vector3 addition of another Vector3's coordinates.
  Vector3& operator+=(const Vector3 &);

  //// Vector3 subtraction of another Vector3's coordinates.
  Vector3& operator-=(const Vector3 &);

  //// Multiplies coordinates by scalar m.
  Vector3& operator*=(const real &m);

  //// Divides coordinates by scalar m.
  Vector3& operator/=(const real &m);

  //// Outputs coordinates delimited by single space.
  friend ostream& operator<<(ostream& s, const Vector3 &v);

  //// Inputs coordinates delimited by single space.
  friend istream& operator>>(istream& s, Vector3 &v);

private:
  real c[3];  /* coordinates */
};

#endif










