// triangle.h: interface for the triangle class.
//
//////////////////////////////////////////////////////////////////////

/* 
CLASS 
  triangle

  Defines a triangle class for easy triangle handling
  (Includes rotation operator (|))2

KEYWORDS
  linear, algebra, triangle, rotation, baricenter

AUTHORS
  Yevgeny Menaker. (genmen@post.tau.ac.il)

  Copyright: SAMBA group, Tel-Aviv Univ. Israel, 2001.

CHANGES LOG
<UL>
<LI>
</UL>

GOALS
  This class simplifies the work with triangle form
  allowing easy and fast operation with them. The core
  of this class is rotation operator, which performs
  3D transformation, rotating one triangle to fit the 
  second one effectively.
  
USAGE
  There are provided three operators: assignment(=), shifting the triangle
  by Vector3 (+) and rotation operator(|).
  In addition the Baricenter method is provided to calculate the 
  baricenter of triangle.

  A new triangle may be constructed with 3 3D vectors - one
  for each vertix of triangle
  EXAMPLE
    Vector3 v_A(1.0, 2.0, 3.0);
	Vector3 v_B(1.0, 5.0, 1.0);
	Vector3 v_C(0.0, 2.5, 2.0);

	triangle tr(v_A, v_B, v_C);

  END

  You may calculate the baricenter of given triangle:

  EXAMPLE
    Vector3 baricenter = tr.Baricenter();
  END

  Shifting of triangle:

  EXAMPLE
	triangle tr2 = tr + Vector3(1.0,1.0,1.0);
  END

  And finally the calculation of 3D transformation
  of two triangles, assuming they can be fit

  EXAMPLE
    RigidTrans3 rt3 = tr | tr2;

    cout << "The rotation is: " << rt3;

  END

*/
#ifndef _TRIANGLE_H
#define _TRIANGLE_H

#include "Vector3.h"
#include "RigidTrans3.h"

class triangle  
{
public:
	// Default constructor
	triangle();
	// Constructor with vertices initialization
	triangle(Vector3& a, Vector3& b, Vector3& c);
	// Copy constructor
	triangle(const triangle& tr);
	virtual ~triangle();

	// Assignment operator
	triangle& operator=(const triangle& tr);

	// Calculates the Baricenter of the triangle
	// We assume that every vertix of triangle has
	// the same weight (1)
	Vector3& Baricenter();

	// Shift the Baricenter of triangle (add vector)
	triangle& operator+(const Vector3& vec);

	// Returns rotation transformation to fit this triangle
	// with triangle tr in 3D space
	// Returns zero rotmatrix if triangles cannot be fit
	// because of any reason
	RigidTrans3& operator|(triangle& tr);



private:
	// Triangle vertices
	Vector3 m_A;
	Vector3 m_B;
	Vector3 m_C;

};

#endif